import argparse
import json
import numbers
import pathlib

import appdirs
import numpy as np

from .. import model
from .. import preproc
from .. import rate

APP_DIR = pathlib.Path(appdirs.user_config_dir(appname="nanite"))
PROFILE_PATH = APP_DIR / "cli_profile.cfg"

DEFAULTS = {"model_key": "sneddon_spher_approx",
            "preprocessing": ["compute_tip_position",
                              "correct_force_offset",
                              "correct_tip_offset"],
            "preprocessing_options": {},
            "range_type": "absolute",
            "range_x": [0, 0],
            "segment": 0,
            "weight_cp": 5e-7,
            "rating regressor": "Extra Trees",
            "rating training set": "zef18",
            }


class JSONPathEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, pathlib.Path):
            return f"{obj}"
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


class Profile:
    def __init__(self, path=PROFILE_PATH, create=True):
        """Initialize settings file (create if it does not exist)"""
        path = pathlib.Path(path)
        if create:
            path.parent.mkdir(exist_ok=True, parents=True)
            path.touch()
        elif not path.exists():
            raise ValueError("Please run `nanite-setup-profile` first!")
        self.path = path
        # initialize with defaults
        for key in DEFAULTS:
            self[key]

    def __getitem__(self, key):
        default = DEFAULTS[key]
        data = self.load()
        val = data.get(key, default)
        # also set default
        self[key] = val
        return val

    def __setitem__(self, key, value):
        if key.startswith("fit param"):
            if not (key.endswith("value")
                    or key.endswith("vary")):
                raise ValueError("Invalid key: '{}'".format(key))
        data = self.load()
        data[key] = value
        self.save(data)

    def set_fit_params(self, params):
        for p in params:
            self["fit param {} value".format(p)] = params[p].value
            self["fit param {} vary".format(p)] = params[p].vary

    def get_fit_params(self):
        cdict = self.load()
        # get model key
        model_key = cdict["model_key"]
        default = model.get_init_parms(model_key)
        for p in default:
            vkey = "fit param {} value".format(p)
            if vkey in cdict:
                default[p].value = cdict[vkey]
            fkey = "fit param {} vary".format(p)
            if fkey in cdict:
                assert isinstance(cdict[fkey], bool)
                default[p].vary = cdict[fkey]

        # write
        self.set_fit_params(default)
        return default

    def load(self):
        """Loads the profile file returning a dictionary"""
        try:
            text = self.path.read_text()
            return json.loads(text)
        except json.decoder.JSONDecodeError:
            return self.load_legacy()

    def load_legacy(self):
        """Load profile from the old profile file format"""
        with self.path.open() as fop:
            fc = fop.readlines()
        cdict = {}
        for line in fc:
            line = line.strip()
            var, val = line.split("=", 1)
            # support "approach" and "retract" from pre 1.8.0 versions
            var = var.strip()
            val = val.strip()
            if var == "segment":
                if val == "approach":
                    val = "0"
                elif val == "retract":
                    val = "1"
            cdict[var] = val

        for key in cdict:
            default = DEFAULTS[key]
            if isinstance(default, list):
                val = cdict[key].split(",")
                if isfloat(val[0]) and isfloat(val[1]):
                    val = [float(vv) for vv in val]
            elif isinstance(default, str):
                val = cdict[key]
            elif isinstance(default, numbers.Integral):
                val = int(cdict[key])
            else:
                val = float(cdict[key])
            cdict[key] = val
        return cdict

    def save(self, cdict):
        """Save a settings dictionary into a file"""
        self.path.write_text(json.dumps(cdict,
                                        indent=2,
                                        sort_keys=True,
                                        ensure_ascii=False,
                                        allow_nan=True,
                                        cls=JSONPathEncoder,
                                        ))


def setup_profile():
    """Help user to create a fitting profile"""
    parser = setup_profile_parser()
    parser.parse_args()

    pf = Profile()

    print("\nDefine preprocessing:")
    steps = [pp.identifier for pp in preproc.PREPROCESSORS]
    cur = pf["preprocessing"]
    curid = ",".join([str(steps.index(cc) + 1) for cc in cur])
    for ii, st in enumerate(steps):
        print("  {}: {}".format(ii+1, st))
    stp = input("(currently '{}'): ".format(curid))
    if stp:
        pf["preprocessing"] = [steps[int(ii) - 1] for ii in stp.split(",")]

    print("\nSelect model number:")
    models = sorted(model.models_available.keys())
    idx = models.index(pf["model_key"])
    for ii, mm in enumerate(models):
        print("  {}: {}".format(ii+1, mm))
    mod = input("(currently '{}'): ".format(idx + 1))
    if mod:
        newmod = models[int(mod) - 1]
        if newmod != pf["model_key"]:
            pf["model_key"] = newmod
            pf["params_initial"] = ""

    print("\nSet fit parameters:")
    params = pf.get_fit_params()
    usedmod = model.models_available[pf["model_key"]]
    for ii, p in enumerate(params):
        unit = usedmod.parameter_units[ii]
        if unit:
            unit = " [{}]".format(unit)
        value = input(
            "- initial value for {}{} (currently '{}'): ".format(
                p, unit, params[p].value))
        if value:
            params[p].value = float(value)
        while True:
            vary = input(
                "  vary {} (currently '{}'): ".format(p, params[p].vary))
            if vary:
                if vary.strip().lower() == "true":
                    vary = True
                elif vary.strip().lower() == "false":
                    vary = False
                else:
                    print("Please type 'true' or 'false'.")
                    continue
                params[p].vary = vary
            break
        pf.set_fit_params(params)

    print("\nSelect range type (absolute or relative):")
    while True:
        rt = input("(currently '{}'): ".format(pf["range_type"]))
        if rt:
            if rt not in ["absolute", "relative"]:
                print("Please choose 'absolute' or 'relative'.")
                continue
            pf["range_type"] = rt
        break

    print("\nSelect fitting interval:")
    ival = np.array(pf["range_x"]) * 1e6
    left = input("left [µm] (currently '{}'): ".format(ival[0]))
    if left:
        ival[0] = float(left)
    right = input("right [µm] (currently '{}'): ".format(ival[1]))
    if left:
        ival[1] = float(right)
    pf["range_x"] = list(ival*1e-6)

    print("\nSuppress residuals near contact point:")
    wcpd = pf["weight_cp"] * 1e6
    wcp = input("size [µm] (currently '{}'): ".format(wcpd))
    if wcp:
        pf["weight_cp"] = float(wcp) * 1e-6

    print("\nSelect training set:")
    cts = pf["rating training set"]
    while True:
        ts = input("training set (path or label) "
                   + "(currently '{}'): ".format(cts))
        if ts:
            # Test whether the training set is implemented in nanite or
            # whether the data exist and are complete.
            ir = rate.IndentationRater
            if not pathlib.Path(ts).exists():
                pp = ir.get_training_set_path(ts)
            else:
                pp = ts
            try:
                ir.load_training_set(path=pp)
            except OSError:
                print("No training set found for '{}'!".format(ts))
                continue
            else:
                pf["rating training set"] = ts
                break
        else:
            break

    print("\nSelect rating regressor:")
    regs = rate.reg_names
    rcur = pf["rating regressor"]
    rcurid = regs.index(rcur) + 1
    for ii, st in enumerate(regs):
        print("  {}: {}".format(ii+1, st))
    rstp = input("(currently '{}'): ".format(rcurid))
    if rstp:
        pf["rating regressor"] = regs[int(rstp) - 1]

    print("\nDone. You may edit all parameters in '{}' ".format(pf.path)
          + "or alternatively run nanite-setup-profile again.")


def setup_profile_parser():
    descr = "Set up a profile for fitting and rating. The profile is stored " \
            + "in the user's default configuration directory. Setting up a " \
            + "profile is required prior to running `nanite-fit` and " \
            + "`nanite-rate`."
    parser = argparse.ArgumentParser(description=descr)
    return parser


def isfloat(value):
    try:
        value = float(value)
    except ValueError:
        return False
    else:
        return True



import inspect

import numpy as np
from scipy import ndimage

#: Valid keyword arguments for feature types
VALID_FEATURE_TYPES = ["all", "binary", "continuous"]


class IndentationFeatures(object):
    def __init__(self, dataset=None):
        #: current dataset from which features are computed
        self.dataset = dataset

    @property
    def is_fitted(self):
        if self.is_valid:
            return self.dataset.fit_properties["success"]
        else:
            return False

    @property
    def is_valid(self):
        return bool(self.dataset.fit_properties)

    @property
    def has_contact_point(self):
        if self.is_fitted:
            pint = self.dataset.fit_properties["params_fitted"]
            return "contact_point" in pint
        else:
            return False

    @property
    def contact_point(self):
        if self.has_contact_point:
            pint = self.dataset.fit_properties["params_fitted"]
            return pint["contact_point"].value
        else:
            raise ValueError("No contact point in data!")

    @property
    def datafit_apr(self):
        seg = self.dataset["segment"] == 0
        f = self.dataset["fit"][seg].copy()
        return f

    @property
    def datares_apr(self):
        return self.datafit_apr - self.datay_apr

    @property
    def datax_apr(self):
        xaxis = self.dataset.fit_properties["x_axis"]
        seg = self.dataset["segment"] == 0
        x = self.dataset[xaxis][seg].copy()
        # Make sure everything is ok
        assert x[0] > x[-1], "Approach from large distances towards lower"
        return x

    @property
    def datay_apr(self):
        yaxis = self.dataset.fit_properties["y_axis"]
        seg = self.dataset["segment"] == 0
        y = self.dataset[yaxis][seg].copy()
        return y

    @property
    def meta(self):
        return self.dataset.metadata

    @staticmethod
    def compute_features(idnt, which_type="all", names=None,
                         ret_names=False):
        """Compute the features for a data set

        Parameters
        ----------
        idnt: nanite.Indentation
            A dataset to rate
        names: list of str
            The names of the rating methods to use,
            e.g. ["rate_apr_bumps", "rate_apr_mon_incr"].
            If None (default), all available rating methods are
            used.

        Notes
        -----
        `names` may include features that are excluded by `which_type`.
        E.g. if a "bool" feature is in `names` but `which_type` is
        "float", then the "bool" feature will be silently ignored.
        """
        inst = IndentationFeatures(idnt)
        # make sure to keep the order of `names`, i.e. only compute
        # names if they are not given.
        if names is None or which_type != "all":
            names = IndentationFeatures.get_feature_names(
                which_type=which_type, names=names)
        samples = []
        for nameii in names:
            sii = float(getattr(inst, nameii)())
            samples.append(sii)
        samples = np.array(samples, dtype=float).flatten()
        if ret_names:
            return samples, names
        else:
            return samples

    @classmethod
    def get_feature_funcs(cls, which_type="all", names=None):
        """Return functions that compute features from a dataset

        Parameters
        ----------
        names: list of str
            The names of the rating methods to use,
            e.g. ["rate_apr_bumps", "rate_apr_mon_incr"].
            If None (default), all available rating methods are
            returned.
        which_type: str
            Which features to return: ["all", "bool", "float"].

        Returns
        -------
        raters: list of tuples (name, callable)
            Each item in the list consists contains the name
            of the rating method and the corresponding rating
            method.
        """
        fnames = cls.get_feature_names(which_type=which_type, names=names)
        ffuncs = [(ff, getattr(cls, ff)) for ff in fnames]
        return ffuncs

    @classmethod
    def get_feature_names(cls, which_type="all", names=None,
                          ret_indices=False):
        """Get features names

        Parameters
        ----------
        which_type: str or list of str
            Return only features that are of a certain type.
            See `VALID_FEATURE_TYPES` for valid strings.
        names: list of str
            Return only features that are in this list.
        ret_indices: bool
            If True, also return the internal feature indices.

        Returns
        -------
        name_list: list of str
            List of feature names (callables of this class)
        """
        if isinstance(which_type, (list, tuple)):
            # recurse
            fnames = []
            for wt in which_type:
                fnames += cls.get_feature_names(which_type=wt)
        else:
            if which_type not in VALID_FEATURE_TYPES:
                msg = "`which_type` must be one of {}, got '{}'".format(
                    VALID_FEATURE_TYPES, which_type)
                raise ValueError(msg)
            if which_type == "all":
                fstart = "feat_"
            elif which_type == "binary":
                fstart = "feat_bin_"
            elif which_type == "continuous":
                fstart = "feat_con_"
            ffuncs = inspect.getmembers(cls, lambda a: (inspect.isroutine(a)))
            fnames = [ff[0] for ff in ffuncs if ff[0].startswith(fstart)]
        # keep only names requested by the user
        if names:
            # convenience: make sure the requested feature names all exists
            refnames = cls.get_feature_names()
            unknown = [nn for nn in names if nn not in refnames]
            if unknown:
                msg = "Unknown feature names: '{}'".format(",".join(unknown))
                raise ValueError(msg)
            # only use selected features
            fnames = [ff for ff in fnames if ff in names]
        fnames = sorted(fnames)
        if ret_indices:
            # return the internal index
            allnames = cls.get_feature_names()
            indices = []
            for ii, aff in enumerate(allnames):
                if aff in fnames:
                    indices.append(ii)
            return fnames, np.array(indices, dtype=int)
        else:
            return fnames

    # for bool features, 1 means good and 0 means bad
    def feat_bin_apr_spikes_count(self):
        """spikes during IDT

        Sudden spikes in indentation curve
        """
        if self.has_contact_point:
            cp = self.contact_point
            # indentation part
            indidx = self.datax_apr < cp
            diff = self.datares_apr[indidx]
            diff = diff[~np.isnan(diff)]
            if len(diff) > 50:
                # find regions with peaks
                diff_smooth = ndimage.gaussian_filter1d(diff, sigma=11)
                delta1 = diff - diff_smooth
                diff_smooth2 = ndimage.gaussian_filter1d(diff, sigma=1)
                delta2 = diff_smooth2 - diff_smooth
                std = np.std(delta1)
                peakarray = np.diff(np.abs(delta2) > 3*std)
                npeaks = np.sum(peakarray)
                value = npeaks <= 5
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_bin_cp_position(self):
        """CP outside of data range

        Contact point position outside of range
        """
        if self.has_contact_point:
            cp = self.contact_point
            x = self.datax_apr
            if cp < np.min(x) or cp > np.max(x):
                value = False
            else:
                value = True
            return value
        else:
            return np.nan

    def feat_bin_size(self):
        """dataset too small

        Number of points in indentation curve
        """
        if self.is_valid:
            a_ind = self.datay_apr
            num = a_ind.shape[0]
            if num < 600:
                value = False
            else:
                value = True
        else:
            value = np.nan
        return value

    def feat_con_apr_flatness(self):
        """flatness of APR residuals

        fraction of the positive-gradient residuals in the approach part
        """
        if self.has_contact_point:
            cp = self.contact_point
            # baseline indices of approach curve
            # (approaches from pos values)
            blidx = self.datax_apr > cp
            # get residuals of approach curve
            r_bl = self.datares_apr[blidx]
            r_bl = r_bl[~np.isnan(r_bl)]
            # perform gaussian blur
            sigma = max(5, (int(r_bl.shape[0]/120)//2)*2+1)
            y = ndimage.gaussian_filter1d(r_bl, sigma=sigma)
            if len(y) > 2:
                grad = np.gradient(y)
                pos = np.sum(grad > 0)
                neg = np.sum(grad < 0)
                value = pos/(pos + neg)
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_apr_sum(self):
        """residuals of APR

        absolute sum of the residuals in the approach part
        """
        if self.has_contact_point:
            cp = self.contact_point
            # indentation part
            indidx = self.datax_apr > cp
            diff = self.datares_apr[indidx]
            diff = diff[~np.isnan(diff)]
            xin = self.datax_apr
            yin = self.datay_apr

            norm = xin.size * np.max(yin)
            value = np.sum(np.abs(diff)) / norm * 100
            value = np.log(1+value)
        else:
            value = np.nan
        return value

    def feat_con_apr_size(self):
        """relative APR size

        length of the approach part relative to the indentation part
        """
        if self.has_contact_point:
            cp = self.contact_point
            # baseline indices of approach curve
            # (approaches from pos values)
            x = self.datax_apr
            aprsize = np.sum(x > cp)
            tolsize = x.shape[0]
            value = 1 - aprsize/tolsize
            return value
        else:
            return np.nan

    def feat_con_bln_slope(self):
        """slope of BLN

        slope obtained from a linear least-squares fit to the baseline
        """
        if self.has_contact_point:
            cp = self.contact_point
            x = self.datax_apr
            y = self.datares_apr
            # break point is between xmax and cp
            breakp = (np.max(x) + cp) / 2
            # left of break point and no nans
            valid = (x > breakp) * (~np.isnan(x))

            # slope range
            x_sl = x[valid]
            y_sl = y[valid]

            if x_sl.size > 20:
                # fit a line to the data
                A = np.vstack([x_sl, np.ones(len(x_sl))]).T
                m, _c = np.linalg.lstsq(A, y_sl, rcond=None)[0]
                # normalize with maximum force
                value = m / np.max(self.datay_apr)
                value = np.log(1 + np.abs(value)) / 10
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_bln_variation(self):
        """variation in BLN

        comparison of the forces at the beginning and at the end
        of the baseline
        """
        if self.has_contact_point:
            cp = self.contact_point
            r_bl = self.datares_apr[self.datax_apr > cp]
            r_bl = r_bl[~np.isnan(r_bl)]
            # skip 10%
            offset = int(r_bl.shape[0]*.1)
            if offset:
                r_bl = r_bl[:-offset]
            if r_bl.shape[0] > 20:
                avg1 = np.average(r_bl[:10])
                avg2 = np.average(r_bl[-10:])
                # normalize with maximum force and multiply by 1000
                maxforce = np.max(self.datay_apr)
                value = np.abs(avg1-avg2)/maxforce * 1e3
                value = np.log(1 + value) / 5
            else:
                value = np.nan
            return value
        else:
            return np.nan

    def feat_con_cp_curvature(self):
        """curvature at CP

        curvature of the force-distance data at the contact point
        """
        if self.has_contact_point:
            cp = self.contact_point
            x = self.datax_apr
            y = self.datay_apr
            # 10 % of indentation is range
            cpid = np.argmin(np.abs(x - cp))
            maxid = np.argmax(y)
            incl = np.abs(maxid - cpid) // 10
            if incl > 5:
                reg = y[cpid - incl:cpid + incl]
                if len(reg):
                    lin = np.linspace(reg.min(), reg.max(), reg.size)
                    diff = np.sum(reg-lin)
                    value = diff / y.max() * 10
                    value = np.log(1 + np.abs(value)) * np.sign(value) / 4
                else:
                    value = np.nan
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_cp_magnitude(self):
        """residuals at CP

        mean value of the residuals around the contact point
        """
        if self.has_contact_point:
            cp = self.contact_point
            # use 10% of the indentation part as size
            r_range = int(np.sum(self.datax_apr < cp) * .1)
            cpidx = np.nanargmin(np.abs(self.datax_apr-cp))
            rat_range = np.abs(
                self.datares_apr[cpidx - r_range:cpidx + r_range])
            # normalization with setpoint of curve gives us a number
            if len(rat_range):
                value = np.nansum(rat_range) / np.max(self.datay_apr) / 100
            else:
                value = np.nan
            return value
        else:
            return np.nan

    def feat_con_idt_maxima_75perc(self):
        """maxima in IDT residuals

        sum of the indentation residuals' maxima in three intervals
        in-between 25% and 100% relative to the maximum indentation
        """
        if self.has_contact_point:
            yin = self.datay_apr
            fit = self.datafit_apr
            xin = self.datax_apr
            cp = self.contact_point
            id100 = np.argmin(xin)
            id000 = np.argmin(np.abs(xin - cp))
            id025 = int(id000 + .25 * (id100 - id000))
            idmin, idmax = min(id025, id100), max(id025, id100)
            if idmin != idmax:
                # find zeros
                idcen = idmin + (idmax - idmin) // 2
                smooth = ndimage.gaussian_filter1d(yin - fit, sigma=11)
                idzero1 = idmin + np.argmin(np.abs(smooth[idmin:idcen]))
                idzero2 = idcen + np.argmin(np.abs(smooth[idcen:idmax]))
                # change of sign in 1st, 2nd, and 3rd part of indentation
                ydiffs = []
                if idmin != idzero1:
                    y1 = np.max(np.abs((yin - fit)[idmin:idzero1]))
                    ydiffs.append(y1)
                if idzero1 != idzero2:
                    y2 = np.max(np.abs((yin - fit)[idzero1:idzero2]))
                    ydiffs.append(y2)
                if idzero2 != idmax:
                    y3 = np.max(np.abs((yin - fit)[idzero2:idmax]))
                    ydiffs.append(y3)
                if ydiffs:
                    # combine the changes (ydiff2 has different sign)
                    ydiff = np.sum(ydiffs)
                    ydiff /= yin.max()
                    value = np.log(1 + ydiff) * 2
                else:
                    value = np.nan
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_idt_monotony(self):
        """monotony of IDT

        change of the gradient in the indentation part
        """
        if self.has_contact_point:
            cp = self.contact_point
            # indentation part
            indidx = self.datax_apr < cp
            # get approach curve
            a_ind = self.datay_apr[indidx]
            # perform gaussian blur
            y = ndimage.gaussian_filter1d(a_ind, sigma=2)
            if len(y) > 2:
                grad = np.gradient(y)
                gz = np.abs(np.sum(grad[grad > 0]))
                lz = np.abs(np.sum(grad[grad < 0]))
                value = np.sum(indidx) * lz / gz
                value = np.log(1 + value) / 10
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_idt_sum(self):
        """overall IDT residuals

        sum of the residuals in the indentation part
        """
        if self.has_contact_point:
            cp = self.contact_point
            x = self.datax_apr
            y = self.datay_apr
            f = self.datafit_apr
            idind = x < cp
            diff = (y - f)[idind]
            if diff.size:
                area = np.nansum(np.abs(diff))/diff.size
                norm = np.abs(y.max() - y.min())/2
                value = area/norm
                value = np.log(1 + value) * 5
            else:
                value = np.nan
        else:
            value = np.nan
        return value

    def feat_con_idt_sum_75perc(self):
        """residuals in 75% IDT

        sum of the residuals in the indentation part in-between
        25% and 100% relative to the maximum indentation
        """
        if self.has_contact_point:
            yin = self.datay_apr
            fit = self.datafit_apr
            xin = self.datax_apr
            cp = self.contact_point
            id100 = np.argmin(xin)
            id000 = np.argmin(np.abs(xin - cp))
            id025 = int(id000 + .25 * (id100 - id000))
            idmin, idmax = min(id025, id100), max(id025, id100)
            ydiff = np.sum(np.abs(yin - fit)[idmin:idmax])
            xnorm = np.abs(xin[idmin] - xin[idmax])
            # area under the curve (normalization with absolute x interval)
            area = ydiff * xnorm
            # normalize by max force and multiply by 1e6 to get values around 1
            value = area / np.max(yin) * 1e6
            value = np.log(1+value) / 8
        else:
            value = np.nan
        return value

    def feat_con_idt_spike_area(self):
        """area of IDT spikes

        area of spikes appearing in the indentation part
        """
        if self.has_contact_point:
            cp = self.contact_point
            # indentation part
            indidx = self.datax_apr < cp
            diff = self.datares_apr[indidx]
            diff = diff[~np.isnan(diff)]
            if len(diff) > 20:
                # find regions with peaks
                diff_smooth = ndimage.gaussian_filter1d(diff, sigma=11)
                delta1 = diff - diff_smooth
                diff_smooth2 = ndimage.gaussian_filter1d(diff, sigma=1)
                delta2 = np.abs(diff_smooth2 - diff_smooth)
                std = np.std(delta1)
                peaks = np.sum(delta2[delta1 > 3*std])
                # add SD to avoid too many zero-valued samples
                value = (std + peaks) / np.max(self.datay_apr)
                value = np.log(1 + value) * 20
            else:
                value = np.nan
        else:
            value = np.nan
        return value



import lmfit
import numpy as np


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E", value=3e3, min=0)
    params.add("R", value=10e-6, min=0, vary=False)
    params.add("nu", value=.5, min=0, max=0.5, vary=False)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def hertz_sneddon_spherical_approx(delta, E, R, nu, contact_point=0,
                                   baseline=0):
    r"""Hertz model for Spherical indenter - approximation

    .. math::

        F = \frac{4}{3} \frac{E}{1-\nu^2} \sqrt{R} \delta^{3/2}
            \left(1
             - \frac{1}{10} \frac{\delta}{R}
             - \frac{1}{840} \left(\frac{\delta}{R}\right)^2
             + \frac{11}{15120} \left(\frac{\delta}{R}\right)^3
             + \frac{1357}{6652800} \left(\frac{\delta}{R}\right)^4
             \right)

    Parameters
    ----------
    delta: 1d ndarray
        Indentation [m]
    E: float
        Young's modulus [N/m²]
    R: float
        Tip radius [m]
    nu: float
        Poisson's ratio
    contact_point: float
        Indentation offset [m]
    baseline: float
        Force offset [N]

    Returns
    -------
    F: float
        Force [N]

    Notes
    -----
    These approximations are made by the Hertz model:

    - The sample is isotropic.
    - The sample is a linear elastic solid.
    - The sample is extended infinitely in one half space.
    - The indenter is not deformable.
    - There are no additional interactions between sample and indenter.

    Additional assumptions:

    - no surface forces

    Truncated power series approximation:

    This model is a truncated power series approximation of
    :ref:`sec_ref_model_sneddon_spher`. The expected error is more
    than four magnitues lower than the signal (see e.g.
    :ref:`example_model_spherical_indenter`). The Bio-AFM analysis software
    by JPK/Bruker uses the same model.

    References
    ----------
    Sneddon (1965) :cite:`Sneddon1965` (equations 6.13 and 6.15),
    Dobler (personal communication, 2018) :cite:`Dobler`
    """
    aa = 4/3 * E/(1-nu**2)*np.sqrt(R)
    root = contact_point-delta
    pos = root > 0
    bb = np.zeros_like(delta)
    bb[pos] = (root[pos])**(3/2)*(
        + 1
        - 1/10*(root[pos]/R)
        - 1/840*(root[pos]/R)**2
        + 11/15120*(root[pos]/R)**3
        + 1357/6652800*(root[pos]/R)**4)
    return aa*bb + baseline


model_doc = hertz_sneddon_spherical_approx.__doc__
model_func = hertz_sneddon_spherical_approx
model_key = "sneddon_spher_approx"
model_name = "spherical indenter (Sneddon, truncated power series)"
parameter_keys = ["E", "R", "nu", "contact_point", "baseline"]
parameter_names = ["Young's Modulus", "Tip Radius",
                   "Poisson's Ratio", "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "m", "", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



import copy
import hashlib
import numbers
import warnings

import lmfit
import numpy as np
import scipy.signal as spsig

from . import model


FP_DEFAULT = dict(model_key="hertz_para",
                  optimal_fit_edelta=False,
                  optimal_fit_num_samples=100,
                  params_initial=None,
                  preprocessing=[],
                  preprocessing_options={},
                  range_type="absolute",
                  range_x=[0, 0],
                  segment=0,
                  weight_cp=1e-6,
                  gcf_k=1.0,
                  x_axis="tip position",
                  y_axis="force",
                  method="leastsq",
                  method_kws={},
                  )

FP_RESULTS = ["chi_sqr",
              "hash",
              "optimal_fit_delta_array",
              "optimal_fit_delta",
              "optimal_fit_E_array",
              "params_fitted",
              "success",
              "xmax",
              "xmin",
              ]


class FitKeyError(BaseException):
    pass


class FitDataError(BaseException):
    pass


class FitWarning(UserWarning):
    pass


class FitProperties(dict):
    """Fit property manager class

    Provide convenient access to fit properties as a dictionary
    and dynamically manage resets due to new initial parameters.

    Dynamic properties include:

    - set "params_initial" to `None` if the "model_key" changes
    - remove all keys except those in `FP_DEFAULT` if a key that is
      in `FP_DEFAULT` changes (All other keys are considered to be
      obsolete fitting results).

    Additional attributes:
    """

    def __setitem__(self, key, value):
        if key == "segment":
            # Since version 1.8.0, nanite uses an integer to identify
            # the segment. Replace the input accordingly for backwards
            # compatibility.
            if value == "approach":
                value = 0
            elif value == "retract":
                value = 1
        # Proceed with normal checks.
        if key in FP_DEFAULT:
            if (key in self
                and key == "params_initial"
                    and self["params_initial"] is not None
                    and value is not None):
                # check for changed initial parameters
                for pp in self["params_initial"]:
                    s1 = self["params_initial"][pp].__getstate__()
                    s2 = value[pp].__getstate__()
                    if s1 != s2:
                        self.reset()
                        break
            elif key in self and self[key] == value:
                # nothing to do, parameters match
                pass
            else:
                if key == "model_key":
                    # Other model has other parameters.
                    self["params_initial"] = None
                elif key == "range_x":
                    if ("optimal_fit_edelta" in self and
                        self["optimal_fit_edelta"] and
                        "range_x" in self and
                            self["range_x"][1] == value[1]):
                        # Ignore changes in range[0]
                        return
                # Trigger `self.reset`
                self.reset()
        elif key not in FP_RESULTS:
            msg = "Key '{}' not in FP_DEFAULT".format(key)
            raise FitKeyError(msg)
        super(FitProperties, self).__setitem__(key, value)

    def reset(self):
        for key in list(self.keys()):
            if key not in FP_DEFAULT:
                self.pop(key)

    def restore(self, props):
        """update the dictionary without removing any keys"""
        for key in props:
            super(FitProperties, self).__setitem__(key, props[key])


class IndentationFitter(object):
    def __init__(self, idnt, **kwargs):
        """Fit force-distance curves

        Parameters
        ----------
        idnt: nanite.indent.Indentation
            The dataset to fit
        model_key: str
            A key referring to a model in
            `nanite.model.models_available`
        params_initial: instance of lmfit.Parameters
            Parameters for fitting. If not given,
            default parameters are used.
        range_x: tuple of 2
            The range for fitting, see `range_type` below.
        range_type: str
            One of:

            - absolute:
                Set the absolute fitting range in values
                given by the `x_axis`.
            - relative cp:
                In some cases it is desired to be able to
                fit a model only up until a certain indentation
                depth (tip position) measured from the contact
                point. Since the contact point is a fit parameter
                as well, this requires a two-pass fitting.
        preprocessing: list of str
            Preprocessing step identifiers
        preprocessing_options: dict of dicts
            Preprocessing keyword arguments of steps (if applicable)
        segment: int
            Segment index (e.g. 0 for approach)
        weight_cp: float
            Weight the contact point region which shows artifacts
            that are difficult to model with e.g. Hertz.
        gcf_k: float
            Geometrical correction factor :math:`k` for non-single-contact
            data. The measured indentation is multiplied by this factor to
            correct for experimental geometries during fitting,
            e.g. ``gcf_k=0.5`` for parallel-place compression.
        optimal_fit_edelta: bool
            Search for the optimal fit by varying the maximal
            indentation depth and determining a plateau in the
            resulting Young's modulus (fitting parameter "E").
        optimal_fit_num_samples: int
            Number of samples to use for searching the optimal fit
        """
        # Get initial fit parameters
        # IMPORTANT:
        # If there are new additions in the default values,
        # make sure to take these into account in `FP_DEFAULT`.
        self.fp = FitProperties(**FP_DEFAULT)

        # Get parameters from dataset
        # (sorted, such that `model_key` is set before `params_initial`)
        for key in sorted(idnt.fit_properties.keys()):
            if key in FP_DEFAULT:
                self.fp[key] = idnt.fit_properties[key]
        # Get parameters from kwargs
        # (sorted, such that `model_key` is set before `params_initial`)
        for key in sorted(self.fp.keys()):
            if key in kwargs:
                if key not in FP_DEFAULT:
                    msg = "Key '{}' not in FP_DEFAULT".format(key)
                    raise FitKeyError(msg)
                self.fp[key] = kwargs[key]
        # Set initial fitting parameters
        if self.fp["params_initial"] is None:
            self.fp["params_initial"] = self.get_initial_parameters(
                idnt=idnt,
                model_key=self.fp["model_key"]
            )

        # Set arrays
        self.segment = idnt["segment"] == self.fp["segment"]
        self.segment.setflags(write=False)

        self.x_axis = idnt[self.fp["x_axis"]]
        self.x_axis.setflags(write=False)

        self.y_axis = idnt[self.fp["y_axis"]]
        self.y_axis.setflags(write=False)

        self.fit_range = np.zeros_like(self.segment)
        self.fit_curve = np.zeros_like(self.y_axis)
        self.fit_residuals = np.zeros_like(self.y_axis)

        # Fitting parameters might be changed due to iterative fitting.
        # Store them in separate variables to prevent resetting the
        # `FitPropreties` instance `self.fp`.
        self.optimal_fit_edelta = self.fp["optimal_fit_edelta"]
        self.range_type = self.fp["range_type"]
        self.range_x = list(self.fp["range_x"])

        self.hash = self._hash()
        self.fp["hash"] = self.hash

        # Perform sanity checks
        if self.fp["range_type"] not in ["absolute", "relative cp"]:
            msg = "`range_type` must be  'absolute' or 'relative cp'!"
            raise FitKeyError(msg)
        if len(self.fp["range_x"]) != 2:
            raise FitKeyError("`range_x` must have length 2!")
        if (np.isnan(self.fp["range_x"][0]) or
                np.isnan(self.fp["range_x"][1])):
            raise FitKeyError("`range_x` must not contain NaN!")
        if not isinstance(self.fp["segment"], numbers.Integral):
            raise FitKeyError(
                f"`segment` must be integer, got '{self.fp['segment']}!")
        if self.fp["model_key"] not in model.models_available:
            msg = "unknown model '{}'".format(self.fp["model_key"])
            raise FitKeyError(msg)

        if self.fp["optimal_fit_edelta"]:
            # Make sure we have emodulus
            if "E" not in self.fp["params_initial"]:
                msg = "Search for optimal fit requires the parameter 'E'!"
                raise FitKeyError(msg)
            # This only works with absolute range
            if self.fp["range_type"] != "absolute":
                msg = "Only absolute range relative to contact point allowed!"
                raise FitKeyError(msg)

        md_key = self.fp["model_key"]
        md = model.models_available[md_key]
        params = md.get_parameter_defaults()
        for p in params:
            msg = "Unknown fitting parameter '{}' for model '{}'!"
            if p not in self.fp["params_initial"]:
                raise FitKeyError(msg.format(p, md_key))

        if self.fp["range_x"][0] > self.fp["range_x"][1]:
            msg = "Fitting range is inverted: {}".format(self.fp["range_x"])
            warnings.warn(msg, FitWarning)

    def compute_emodulus_vs_mindelta(self, callback=None):
        """Compute elastic modulus vs. minimal indentation curve"""
        segid = self.segment
        xseg = self.x_axis[segid]
        yseg = self.y_axis[segid]

        xmax = np.max(self.fp["range_x"])
        if np.isinf(xmax):
            xmax = np.max(xseg)
        # Disable and remember `optimal_fit_edelta`
        optimal_fit_edelta = self.optimal_fit_edelta
        self.optimal_fit_edelta = False

        # We are agnostic concerning the direction of indentation.
        # `xseg` should start at the baseline.
        seems_approach = np.average(yseg[:10]) < np.average(yseg[-10:])
        if seems_approach and self.fp["segment"] == 0:
            if xseg[0] < xseg[-1]:
                msg = "Unexpected trend in approach x data!"
                raise FitDataError(msg)
        elif seems_approach:
            msg = "Data appears to be 'approach', but is 'retract'!"
            raise FitDataError(msg)
        elif not seems_approach and self.fp["segment"] > 0:
            if xseg[0] > xseg[-1]:
                msg = "Unexpected trend in retract x data!"
                raise FitDataError(msg)
            msg = "Unexpected trend in retract curve!"
            raise FitDataError(msg)
        elif not seems_approach:
            msg = "Data appears to be 'retract', but is 'approach'!"
            raise FitDataError(msg)

        # Fit the range of parameters
        xmin = xseg.min()
        if xmin >= 0:
            msg = "No negative values (indentation) found! " \
                  + "Did you correct for tip offset?"
            raise FitKeyError(msg)
        num_samp = self.fp["optimal_fit_num_samples"]
        indentations = np.linspace(xmin, xmin*.05, num_samp)
        emoduli = np.zeros_like(indentations)
        for ii, x0 in enumerate(indentations):
            # Perform a fit and record the elastic modulus
            self.range_x = [x0, xmax]
            self.fit()
            emoduli[ii] = self.fp["params_fitted"]["E"].value
            if callback and ii % 5 == 0:
                callback(emoduli, indentations)

        self.optimal_fit_edelta = optimal_fit_edelta

        return emoduli, indentations

    @staticmethod
    def compute_opt_mindelta(emoduli, indentations):
        """Determine the plateau of an emodulus-indentation curve

        The following procedure is performed:

        1. Smooth the emodulus data with a Butterworth filter
        2. Label sequences that have similar values by binning
           into ten regions between the min and max.
        3. Ignore sequences with emodulus that is smaller than
           the binning size.
        4. Determine the longest sequence.
        """
        # Perform smoothing of the curve
        # First, perform filtering with butterworth filter
        nb = 1      # Filter order
        wb = 0.05    # Cutoff frequency
        b, a = spsig.butter(nb, wb, output='ba')
        smooth_e = spsig.filtfilt(b, a, emoduli)
        # Second, determine the longest sequence of values
        # that have the same smoothed value.
        ni = 10
        ivals, istep = np.linspace(smooth_e.min(), smooth_e.max(), ni,
                                   endpoint=False, retstep=True)
        ivals += istep/2
        labelarray = np.zeros_like(smooth_e, dtype=int)
        valarray = np.zeros_like(smooth_e, dtype=int)
        # label each sequence with an individual `idx`
        for ii in range(smooth_e.shape[0]):
            valid = np.argmin(np.abs(ivals - smooth_e[ii]))
            if ii == 0:
                idx = 0
            elif valid == valarray[ii-1]:
                pass
            else:
                idx += 1
            labelarray[ii] = idx
            valarray[ii] = valid
        # Determine the longest sequence
        counts = list(np.bincount(labelarray))
        # Ignore values that are below cutoff

        for ii in range(len(counts)):
            labmax = np.argmax(counts)
            labid = np.where(labelarray == labmax)[0][0]
            valmax = ivals[valarray[labid]]
            if valmax > istep:
                break
            counts.pop(labmax)
        else:
            # Nothing found: select the middle value
            warnings.warn("Could not find correct plateau.", FitWarning)
            labmax = 5
        # Determine the interval in the original array
        indices = np.where(labelarray == labmax)[0]
        if len(indices) == 1:
            dopt = indentations[indices[0]]
        else:
            # compute optimal indentation as center
            dopt = np.average(indentations[indices[0]:indices[-1]])
        return dopt

    def _fit(self):
        """Fit a model to the data

        Notes
        -----
        This method is private because it requires the array
        `self.fit_range` before the actual fitting.
        """
        model_key = self.fp["model_key"]
        params_initial = self.fp["params_initial"]
        # modify contact point with gcf_k
        cpi = params_initial["contact_point"].value
        params_initial["contact_point"].set(value=cpi * self.fp["gcf_k"])
        weight_cp = self.fp["weight_cp"]

        # boolean array indexing the segment
        segid = self.segment
        # x: the entire segment (correct with gcf_k)
        xseg = self.x_axis[segid] * self.fp["gcf_k"]
        # y: the entire segment
        yseg = self.y_axis[segid]
        # x: the values being fitted (correct with gcf_k)
        x = self.x_axis[self.fit_range] * self.fp["gcf_k"]
        # y: the values being fitted
        y = self.y_axis[self.fit_range]

        md = model.models_available[model_key]

        # short reference for better readability
        fit_cur = self.fit_curve
        fit_res = self.fit_residuals

        # reset values to nan
        fit_cur[:] = np.nan
        fit_res[:] = np.nan

        # Make sure that we can actually fit by comparing variable fitting
        # parameters and size of x.
        npvaried = np.sum([p[1].vary for p in list(params_initial.items())])
        if npvaried < x.shape[0] - 1:
            # perform fit
            fit = lmfit.minimize(
                fcn=md.residual,
                params=params_initial,
                method=self.fp["method"],
                args=(x, y, weight_cp),
                **self.fp["method_kws"],
                )
            # fitted method
            fit_cur[segid] = md.model(fit.params, xseg)
            # residuals
            fit_res[segid] = md.residual(fit.params, xseg, yseg, weight_cp)
            # inverse contact point correction with gcf_k
            cpf = fit.params["contact_point"].value
            fit.params["contact_point"].set(value=cpf / self.fp["gcf_k"])
            # add fit results to fp dictionary
            self.fp.update({"params_fitted": fit.params,
                            "chi_sqr": fit.chisqr,
                            "xmin": x.min() / self.fp["gcf_k"],
                            "xmax": x.max() / self.fp["gcf_k"],
                            "success": True,
                            })
        else:
            self.fp["success"] = False

    def fit(self):
        """Fit the approach-retract data to a model function
        """
        range_type = self.range_type
        range_x = copy.copy(self.range_x)
        # Set to True in `self._fit`
        self.fp["success"] = False

        if self.optimal_fit_edelta:
            # emodulus-indentation curve:
            emoduli, indentations = self.compute_emodulus_vs_mindelta()
            # determine optimal x0
            dopt = self.compute_opt_mindelta(emoduli, indentations)
            # add indentations/emoduli to self.fp
            self.fp["optimal_fit_E_array"] = emoduli
            self.fp["optimal_fit_delta_array"] = indentations
            self.fp["optimal_fit_delta"] = dopt
            # update fitting parameters for final fit
            self.range_x = [dopt, np.max(self.fp["range_x"])]
            # perform final fit
            self.optimal_fit_edelta = False
            self.fit()
            self.optimal_fit_edelta = True

        elif self.range_type == "absolute":
            # This is easy. Simply set the boolean array of fitting values
            # Exclude data points from other segment
            if range_x[0] != range_x[1]:
                x_data = self.x_axis.copy()
                range_bool = self.segment.copy()
                rmin, rmax = np.min(range_x), np.max(range_x)
                range_bool[x_data < rmin] = False
                range_bool[x_data > rmax] = False
            else:
                range_bool = self.segment
            self.fit_range[:] = range_bool
            self._fit()

        elif range_type == "relative cp":
            # Let's try to solve this with four passes
            # First, get the approximate contact point
            self.range_type = "absolute"
            # Set full range to get estimate of cp
            self.range_x = [0, 0]
            self.fit()
            # Do the following three-times.
            for _i in range(3):
                # get the fitted contact point
                cp = self.fp["params_fitted"]["contact_point"].value
                self.range_x = list(np.array(range_x)+cp)
                # Second, fit with the new contact point as range parameters
                self.fit()

        # Reset range data to original values
        self.range_type = range_type
        self.range_x = range_x

    def get_initial_parameters(self, idnt=None,
                               model_key=FP_DEFAULT["model_key"]):
        """Get initial fit parameters for a specific model

        Parameters
        ----------

        """
        if (model_key == self.fp["model_key"] and
                self.fp["params_initial"] is not None):
            params = self.fp["params_initial"]
        else:
            params = guess_initial_parameters(idnt=idnt, model_key=model_key)
        return params

    def _hash(self):
        """Compute hash identifier for current fit

        The hash is computed without applying the fit, making
        it suitable for caching fits.
        """
        hashlist = []
        # preprocessing
        hashlist.append(self.fp["preprocessing"])
        hashlist.append(self.fp["preprocessing_options"])
        # axes data
        hashlist.append(self.x_axis)
        hashlist.append(self.y_axis)
        # fit parameters
        for key in FP_DEFAULT:
            if (key == "range_x" and
                    self.fp["optimal_fit_edelta"]):
                # range only partly if "optimal_fit_edelta" is True
                hashlist.append(self.fp["range_x"][1])
            elif (key == "optimal_fit_num_samples" and
                  not self.fp["optimal_fit_edelta"]):
                # ignore number of samples if optimal fit is not used
                pass
            else:
                hashlist.append(self.fp[key])
        # join and hash
        myhash = hashlib.md5(obj2bytes(hashlist)).hexdigest()
        return myhash


def guess_initial_parameters(idnt=None,
                             model_key=FP_DEFAULT["model_key"],
                             common_ancillaries=True,
                             model_ancillaries=True):
    """Guess initial fitting parameters

    Parameters
    ----------
    idnt: nanite.indent.Indentation
        The dataset to use for guessing initial fitting parameters
        using ancillary parameters
    model_key: str
        The model key
    common_ancillaries: bool
        Guess global ancillary parameters (such as contact point)
    model_ancillaries: bool
        Use model-related ancillary parameters
    """
    md = model.models_available[model_key]
    params = md.get_parameter_defaults()
    if common_ancillaries and idnt is not None:
        # Guess initial contact point from actual tip position
        # (see `self.compute_tip_position`)
        # Depending on how the current dataset was pre-processed,
        # the column "tip position" might already be corrected by
        # an estimated contact point offset.
        # (see `self.compute_tip_offset`)
        if "tip position" in idnt:
            cpid = idnt.estimate_contact_point_index()
            cp = idnt["tip position"][cpid]
            params["contact_point"].set(cp)
        else:
            msg = "Cannot estimate contact point, because of missing "\
                  + "column 'tip position'"
            warnings.warn(msg, FitWarning)
    if model_ancillaries and idnt is not None:
        anc_dict = idnt.get_ancillary_parameters()
        # set the parameter values
        for anckey in anc_dict:
            if anckey in params:
                if not np.isnan(anc_dict[anckey]):  # ignore nans
                    params[anckey].set(value=anc_dict[anckey])
    return params


def obj2bytes(obj):
    """Bytes representation of an object for hashing"""
    if isinstance(obj, str):
        return obj.encode("utf-8")
    elif isinstance(obj, (bool, int, float, np.bool_)):
        return str(float(obj)).encode("utf-8")
    elif obj is None:
        return b"none"
    elif isinstance(obj, np.ndarray):
        return obj.tobytes()
    elif isinstance(obj, tuple):
        return obj2bytes(list(obj))
    elif isinstance(obj, list):
        return b"".join(obj2bytes(o) for o in obj)
    elif isinstance(obj, dict):
        return obj2bytes(sorted(obj.items()))
    elif isinstance(obj, lmfit.parameter.Parameter):
        return obj2bytes([obj.value, obj.max, obj.min, obj.vary,
                          obj.expr, obj.name])
    else:
        raise ValueError("No rule to convert object '{}' to string.".
                         format(obj.__class__))



"""Test of data set functionalities"""
import copy
import pathlib

from nanite.fit import FitProperties, FP_DEFAULT, FitKeyError
from nanite import IndentationGroup
from nanite import model


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_changed_fit_properties():
    ar = IndentationGroup(jpkfile)[0]
    # Initially, fit properties are not set
    assert not ar.fit_properties
    assert isinstance(ar.fit_properties, dict)
    # Prepprocessing
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_tip_offset",
                            "correct_force_offset"])
    ar.fit_model()
    hash1 = ar.fit_properties["hash"]
    pinit = copy.deepcopy(ar.fit_properties["params_initial"])
    pinit["E"].vary = False
    assert "hash" in ar.fit_properties, "make sure we didn't change anything"
    ar.fit_properties["params_initial"] = pinit
    assert "hash" not in ar.fit_properties
    ar.fit_model()
    assert hash1 != ar.fit_properties["hash"]


def test_fp_reset():
    fp = FitProperties(**FP_DEFAULT)
    fp["weight_cp"] = 1
    assert "weight_cp" in fp
    # Adding other than default keys is possible.
    fp["hash"] = True
    assert "hash" in fp
    # Updating a default key with same value does not
    # reset thedict.
    fp["weight_cp"] = 1
    assert "hash" in fp
    # Changing the default value removes all other keys.
    fp["hash"] = "laskdlai"
    fp["weight_cp"] = 2
    assert "test" not in fp
    assert "rhababer" not in fp
    # test manual resetting
    fp["hash"] = "2"
    fp.reset()
    assert "hash" not in fp


def test_with_dataset():
    ar = IndentationGroup(jpkfile)[0]
    # Initially, fit properties are not set
    assert not ar.fit_properties
    assert isinstance(ar.fit_properties, dict)
    # Prepprocessing
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_tip_offset",
                            "correct_force_offset"])

    # Fitting the data set will populate the dict
    ar.fit_model()
    assert isinstance(ar.fit_properties, dict)
    assert isinstance(ar.fit_properties, FitProperties)
    assert "hash" in ar.fit_properties
    assert ar.fit_properties["success"]
    hash1 = ar.fit_properties["hash"]
    # Change something
    ar.fit_properties["weight_cp"] = 0
    assert "hash" not in ar.fit_properties
    ar.fit_model()
    hash2 = ar.fit_properties["hash"]
    assert hash1 != hash2
    # Change it back
    ar.fit_properties["weight_cp"] = FP_DEFAULT["weight_cp"]
    ar.fit_model()
    hash3 = ar.fit_properties["hash"]
    assert hash1 == hash3


def test_change_model_key():
    ar = IndentationGroup(jpkfile)[0]
    # Prepprocessing
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_tip_offset",
                            "correct_force_offset"])

    # Fitting the data set will populate the dict
    ar.fit_model(model_key="hertz_para")
    # Change the model
    assert ar.fit_properties["params_initial"] is not None
    ar.fit_properties["model_key"] = "hertz_cone"
    # Changing the model key should reset the initial parameters
    assert ar.fit_properties["params_initial"] is None


def test_wrong_key():
    fp = FitProperties(**FP_DEFAULT)
    # unknown properties raise KeyError
    try:
        fp["dolce"] = False
    except FitKeyError:
        pass
    else:
        raise ValueError("Should not be able to set unknown key!")


def test_wrong_params_initial():
    ar = IndentationGroup(jpkfile)[0]
    # Prepprocessing
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_tip_offset",
                            "correct_force_offset"])
    md = model.models_available["hertz_para"]
    params = md.get_parameter_defaults()
    ar.fit_properties["model_key"] = "hertz_cone"
    try:
        ar.fit_model(params_initial=params)
    except FitKeyError:
        # We forced the wrong fitting parameters.
        pass
    else:
        raise ValueError("Should not be able to use wrong fit parameters!")


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test of data set functionalities"""
import pathlib
import time

import numpy as np

from nanite import IndentationGroup
from nanite.fit import FitDataError


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"
badjpk = data_path / "fmt-jpk-fd_single_bad_GWAT_2017-10-17.jpk-force"


def test_emodulus_search():
    ds = IndentationGroup(jpkfile)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"])
    ar.fit_model(model_key="hertz_cone",
                 params_initial=None,
                 x_axis="tip position",
                 y_axis="force",
                 weight_cp=False,
                 segment="approach",
                 optimal_fit_edelta=True,
                 )
    assert "optimal_fit_delta_array" in ar.fit_properties
    assert "optimal_fit_E_array" in ar.fit_properties
    assert "optimal_fit_delta" in ar.fit_properties

    # This assertion might fail when the preprocessing changes
    # or when the search algorithm for the optimal fit changes.
    dopt = -1.2047321672207138e-07
    assert np.allclose(ar.fit_properties["optimal_fit_delta"], dopt)

    if __name__ == "__main__":
        import matplotlib.pylab as plt
        _fig, axes = plt.subplots(2, 1)
        axes[0].plot(ar["tip position"]*1e6, ar["force"]*1e9, label="data")
        axes[0].plot(ar["tip position"]*1e6, ar["fit"]*1e9, label="fit")
        axes[0].legend()
        axes[0].grid()
        axes[0].set_xlabel("indentation [µm]")
        axes[0].set_ylabel("force [nN]")

        deltas = ar.fit_properties["optimal_fit_delta_array"]*1e6
        emod = ar.fit_properties["optimal_fit_E_array"]
        axes[1].plot(deltas, emod, label="emodulus/minimal-indentation-curve")
        axes[1].set_xlabel("minimal indentation [µm]")
        axes[1].set_ylabel("emodulus [Pa]")
        axes[1].vlines(ar.fit_properties["optimal_fit_delta"]*1e6,
                       emod.min(), emod.max(),
                       label="optimal minimal indentation at plateau",
                       color="r")
        axes[1].grid()
        axes[1].legend()
        plt.tight_layout()
        plt.show()


def test_cache_emodulus():
    # Check that the fitting procedure does not perform unneccessary
    # double fits and uses the cached variables for emoduli and
    # minimal indentations.
    ds = IndentationGroup(jpkfile)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"])
    t01 = time.perf_counter()
    ar.fit_model(model_key="hertz_cone",
                 params_initial=None,
                 x_axis="tip position",
                 y_axis="force",
                 weight_cp=False,
                 segment="approach",
                 optimal_fit_edelta=True,
                 )
    t02 = time.perf_counter()

    t11 = time.perf_counter()
    ar.fit_model()
    t12 = time.perf_counter()
    assert (t02-t01)*1e-4 > t12 - \
        t11, "Second computation should yield cached value (faster)!"

    # Make sure that changing the fit model is longer again
    t21 = time.perf_counter()
    ar.fit_model(model_key="hertz_para")
    t22 = time.perf_counter()

    assert ((t22-t21)*1e-4
            > (t12 - t11)), "Changing model_key should slow down computation!"


def test_fit_data_error():
    # a FitDataError is raised when it is not possible to compute an
    # E(delta) curve:
    ds = IndentationGroup(badjpk)
    ar = ds[0]
    ar.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"])
    try:
        ar.compute_emodulus_mindelta()
    except FitDataError:
        pass
    else:
        assert False, "Invalid input data should not allow E(delt) computation"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
import shutil
import tempfile

from nanite.cli import profile


data_path = pathlib.Path(__file__).parent / "data"


def test_profile_getter():
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_profile_")
    name = pathlib.Path(name)
    pf = profile.Profile(path=name)
    # sanity checks (run twice to trigger loading and saving)
    assert pf["segment"] == 0
    assert pf["segment"] == 0
    assert pf["preprocessing"] == ["compute_tip_position",
                                   "correct_force_offset",
                                   "correct_tip_offset"]
    assert pf["preprocessing"] == ["compute_tip_position",
                                   "correct_force_offset",
                                   "correct_tip_offset"]
    pf["preprocessing"] = ["compute_tip_position"]
    assert pf["preprocessing"] == ["compute_tip_position"]
    assert pf["range_x"] == [0, 0]
    assert pf["range_x"] == [0, 0]
    assert pf["weight_cp"] == 5e-7
    assert pf["weight_cp"] == 5e-7


def test_profile_getter_1_7_8():
    """Load a profile from version 1.7.8"""
    tdir = pathlib.Path(tempfile.mkdtemp(prefix="cli_profile_"))
    name = "cli-profile-1.7.8.cfg"
    cfgpath = tdir / name
    shutil.copy2(data_path / name, cfgpath)
    pf = profile.Profile(path=cfgpath)
    assert pf["segment"] == 0
    assert pf["segment"] == 0


def test_profile_getter_2_1_0():
    """Load a profile from version 2.1.0"""
    tdir = pathlib.Path(tempfile.mkdtemp(prefix="cli_profile_"))
    name = "cli-profile-2.1.0.cfg"
    cfgpath = tdir / name
    shutil.copy2(data_path / name, cfgpath)
    pf = profile.Profile(path=cfgpath)
    assert pf["segment"] == 0
    assert pf["segment"] == 0
    assert pf["rating training set"] == "zef18"


def test_profile_fitparams():
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_profile_")
    name = pathlib.Path(name)
    pf = profile.Profile(path=name)
    # sanity checks (run twice to trigger loading and saving)
    pf["model_key"] = "hertz_cone"
    params = pf.get_fit_params()
    params2 = pf.get_fit_params()
    assert params == params2
    assert len(params) == 5
    assert "E" in params.keys()


def test_single_fitparam():
    _, name = tempfile.mkstemp(suffix=".cfg", prefix="test_nanite_profile_")
    name = pathlib.Path(name)
    pf = profile.Profile(path=name)
    # sanity checks (run twice to trigger loading and saving)
    pf["fit param E value"] = 50
    pf["fit param R value"] = 16e-6
    params = pf.get_fit_params()
    assert params["E"].value == 50
    assert params["R"].value == 16e-6


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import pathlib
from pkg_resources import resource_filename

import numpy as np
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import StandardScaler
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.preprocessing import FunctionTransformer


from .features import IndentationFeatures
from .regressors import reg_dict, reg_names, reg_trees


class IndentationRater(IndentationFeatures):
    def __init__(self, regressor=None, scale=None, lda=None,
                 training_set=None, names=None,
                 weight=True, sample_weight=None,
                 *args, **kwargs):
        """Rate quality

        Parameters
        ----------
        regressor: sciki-learn RegressorMixin
            The regressor used for rating
        scale: bool
            If True, apply a Standard Scaler. If a regressor based on
            decision trees is used, the Standard Scaler is not used
            by default, otherwise it is.
        lda: bool
            If True, apply a Linear Discriminant Analysis (LDA). If a
            regressor based on a decision tree is used, LDA is not
            used by default, otherwise it is.
        training_set: tuple of (X, y)
            The training set (samples, response)
        names: list of str
            Feature names to use
        weight: bool
            Weight the input samples by the number of occurrences
            or with `sample_weight`. For tree-based classifiers, set this
            to True to avoid bias.
        sample_weight: list-like
            The sample weights. If set to `None` sample weights
            are computed from the training set.
        *args: list
            Positional arguments for :class:`IndentationFeatures`
        **kwargs:
            Keyword arguments for :class:`IndentationFeatures`

        See Also
        --------
        sklearn.preprocessing.StandardScaler:
            Standard scaler
        sklearn.discriminant_analysis.LinearDiscriminantAnalysis:
            Linear discriminant analysis
        nanite.rate.regressors.reg_trees:
            List of regressors that are identified as tree-based
        """
        if regressor is not None:
            _name = regressor.__class__.__name__
            if lda is None:
                lda = False if _name in reg_trees else True
            if scale is None:
                scale = False if _name in reg_trees else True

        # training set
        if training_set is None:
            # default
            training_set = self.load_training_set(names=names)
        # sample weights
        if sample_weight is None:
            sample_weight = self.compute_sample_weight(*training_set)
        steps = []

        # scaling (does not affect decision trees / random forests)
        if scale:
            steps.append(StandardScaler())
        # linear discriminant analysis
        if lda:
            steps.append(LinearDiscriminantAnalysis())

        if regressor is not None:
            steps.append(regressor)

        if len(steps) == 0:
            dummy = FunctionTransformer(lambda x: x)
            steps.append(dummy)

        #: sklearn pipeline with transforms (and regressor if given)
        self.pipeline = make_pipeline(*steps)

        fit_params = {}
        if regressor is not None and weight:
            # set weighting for regressor
            key = "{}__sample_weight".format(self.pipeline.steps[-1][0])
            fit_params[key] = sample_weight

        if regressor is not None:
            self.pipeline.fit(*training_set, **fit_params)

        names = self.get_feature_names(names=names, which_type="all")

        #: feature names used by the regressor pipeline
        self.names = sorted(names)
        super(IndentationRater, self).__init__(*args, **kwargs)

    def _pre_rate(self, bsample):
        """exclude based on boolean training set"""
        if np.sum(bsample == 0):
            # bad curve
            return False
        else:
            # good curve
            return True

    def _rate(self, sample):
        gd = self.pipeline.predict(np.atleast_2d(sample))
        return gd[0]

    @staticmethod
    def compute_sample_weight(X, y):
        """Weight samples according to occurrence in y"""
        if not np.all(np.array(y, dtype=int) == y):
            msg = "Only integer ratings allowed."
            raise NotImplementedError(msg)
        weight = np.zeros(y.shape[0], dtype=float)
        for ii in range(11):
            idxii = y == ii
            occur = np.sum(idxii)
            if occur:
                # Sometimes the training set is not large enough.
                # If no occurences were found, the weights remain
                # zero.
                weight[idxii] = 1 / occur
        # normalize
        weight /= np.sum(weight)
        return weight

    @staticmethod
    def get_training_set_path(label="zef18"):
        """Return the path to a training set shipped with nanite

        Training sets are stored in the `nanite.rate`
        module path with ``ts_`` prepended to `label`.
        """
        data_loc = "nanite.rate"
        resp_path = resource_filename(data_loc, "ts_{}".format(label))
        return resp_path

    @classmethod
    def load_training_set(cls, path=None, names=None,
                          which_type=["continuous"],
                          remove_nan=True, ret_names=False):
        """Load a training set from a directory

        By default, only the "continuous" features are imported. The
        "binary" features are not needed for training; they are used
        to sort out new force-distance data.
        """
        fnames = cls.get_feature_names(which_type=which_type, names=names)
        sample_paths = []
        if path is None:
            path = cls.get_training_set_path()
        path = pathlib.Path(path).resolve()

        resp_path = str(path / "train_response.txt")
        for fn in fnames:
            resf = str(path / "train_{}.txt".format(fn))
            sample_paths.append(resf)

        samples = [np.loadtxt(sp, dtype=float, ndmin=2) for sp in sample_paths]
        samples = np.concatenate(samples, axis=1)
        response = np.loadtxt(resp_path, dtype=float)
        if remove_nan:
            # Remove nan-values from training set
            valid = ~np.isnan(np.sum(samples, axis=1))
            samples = samples[valid, :]
            # remove corresponding responses
            response = response[valid]

        res = [samples, response]

        if ret_names:
            res.append(fnames)

        return res

    def rate(self, samples=None, datasets=None):
        """Perform rating step

        Parameters
        ----------
        samples: 1d or 2d ndarray (cast to 2d ndarray) or None
            Measured samples, if set to None, `dataset` must be given.
        dataset: list of nanite.Indentation
            Full, fitted measurement

        Returns
        -------
        ratings: list
            Resulting ratings
        """
        if samples is None and datasets is None:
            # use dataset from IndentationFeature
            datasets = [self.dataset]
        elif datasets is None:
            # distinguish between binary and other samples
            fsamples = []
            bsamples = []
            fnames = self.get_feature_names(
                names=self.names,
                which_type=["continuous"])
            for samp in samples:
                fsamp = []  # continuous samples
                bsamp = []  # binary samples
                for ii, name in enumerate(self.names):
                    if name in fnames:
                        fsamp.append(samp[ii])
                    else:
                        assert name.startswith("feat_bin_")
                        bsamp.append(samp[ii])
                fsamples.append(fsamp)
                bsamples.append(bsamp)
        else:
            if not isinstance(datasets, (list, tuple)):
                datasets = [datasets]
            fsamples = []
            bsamples = []
            # continuous features
            for idnt in datasets:
                samp = self.compute_features(
                    idnt=idnt,
                    names=self.names,
                    which_type=["continuous"])
                fsamples.append(samp)
            # binary features
            for idnt in datasets:
                bsamp = self.compute_features(idnt=idnt,
                                              names=self.names,
                                              which_type="binary")
                bsamples.append(bsamp)

        fsamples = np.atleast_2d(fsamples)
        bsamples = np.atleast_2d(bsamples)

        ratings = []
        for bsamp, fsamp in zip(bsamples, fsamples):
            if not self._pre_rate(bsamp):
                # certainly a bad curve
                gd = 0
            elif np.isnan(np.sum(fsamp)):
                # ignore nan-valued samples
                gd = -1
            else:
                gd = self._rate(fsamp)
            ratings.append(gd)

        return np.array(ratings).flatten()


def get_available_training_sets():
    """List of internal training sets"""
    data_loc = "nanite"
    resp_path = resource_filename(data_loc, "rate")
    avail = []
    for pp in pathlib.Path(resp_path).glob("ts_*"):
        avail.append(pp.name[3:])
    return sorted(avail)


def get_rater(regressor, training_set="zef18", names=None,
              lda=None, **reg_kwargs):
    """Convenience method to get a rater

    Parameters
    ----------
    regressor: str or RegressorMixin
        If a string, must be in `reg_names`.
    training_set: str or pathlib.Path or tuple (X, y)
        A string label representing a training set shipped with
        nanite, the path to a training set, or a tuple
        representing the training set (samples, response)
        for use with sklearn.
    names: list of str
        Only use these features for rating
    lda: bool
        Perform linear discriminant analysis

    Returns
    -------
    irater: nanite.IndentationRater
        The rating instance.
    """
    avr = get_available_training_sets()
    if isinstance(training_set, tuple):
        pass
    else:
        if training_set in avr:
            ts_path = IndentationRater.get_training_set_path(
                label=training_set)
        else:
            ts_path = training_set
        training_set = IndentationRater.load_training_set(
            path=ts_path,
            names=names)

    if len(training_set) != 2:
        raise ValueError("Expected training_set of the form (X, y)!")

    if isinstance(regressor, str):
        if regressor not in reg_names:
            msg = "Unknown regressor name: '{}'!".format(regressor) \
                  + " Please pass your own sklearn RegressorMixin."
            raise ValueError(msg)
        reg_cl, default_kw = reg_dict[regressor]
        kw = default_kw.copy()
        kw.update(reg_kwargs)
        regr = reg_cl(**kw)
    else:
        regr = regressor

    rater = IndentationRater(regressor=regr,
                             training_set=training_set,
                             names=names,
                             lda=lda)
    return rater



"""Test input/output of user rating data"""
import pathlib
import shutil
import tempfile

import h5py
import numpy as np
import pytest

from nanite import model, IndentationGroup
from nanite.rate.io import RateManager, hdf5_rated, load_hdf5, save_hdf5
from nanite.rate.rater import IndentationRater

data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def setuph5(ret_idnt=False, path=jpkfile):
    tdir = tempfile.mkdtemp(prefix="test_nanite_rate_io_")
    tdir = pathlib.Path(tdir)
    h5path = tdir / "simple.h5"
    grp = IndentationGroup(path)
    for idnt in grp:
        idnt.apply_preprocessing(["compute_tip_position"])

        inparams = model.model_hertz_paraboloidal.get_parameter_defaults()
        inparams["baseline"].vary = True
        inparams["contact_point"].set(1.8e-5)

        # Fit with absolute full range
        idnt.fit_model(model_key="hertz_para",
                       params_initial=inparams,
                       range_x=(0, 0),
                       range_type="absolute",
                       x_axis="tip position",
                       y_axis="force",
                       segment="approach",
                       weight_cp=False)

        save_hdf5(h5path=h5path,
                  indent=idnt,
                  user_rate=5,
                  user_name="hans",
                  user_comment="this is a comment",
                  h5mode="a")
    else:
        idnt = grp[0]

    if ret_idnt:
        return tdir, h5path, idnt
    else:
        return tdir, h5path


def test_hdf5rated():
    tdir, h5path, idnt = setuph5(ret_idnt=True)
    is_rated, rating, comment = hdf5_rated(h5path, idnt)
    assert is_rated
    assert rating == 5
    assert comment == "this is a comment"


def test_rate_manager_basic():
    tdir, h5path, idnt = setuph5(ret_idnt=True)
    rmg = RateManager(h5path)
    # sanity checks
    rr = rmg.ratings[0]
    assert rr["name"] == "hans"
    assert rr["rating"] == 5
    # file name preserved
    ds = rmg.datasets[0]
    assert jpkfile.name in ds.path.name
    # features are the same
    idr = IndentationRater
    ss = rmg.samples[0]
    assert np.allclose(ss, idr.compute_features(idnt))
    # rates
    assert np.ndarray.item(rmg.get_rates(which="user")) == 5
    # This will fail when the hyper-parameters for "Extra Trees" change
    # or when new features are added.
    assert np.allclose(
        np.ndarray.item(rmg.get_rates(which="Extra Trees",
                                      training_set="zef18")),
        3.5492840783289035)


def test_rate_manager_crossval():
    path = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"
    tdir, h5path = setuph5(path=path)
    rmg = RateManager(h5path)
    cv = rmg.get_cross_validation_score(regressor="Extra Trees",
                                        training_set=None,
                                        n_splits=2,
                                        random_state=42)
    assert np.all(cv == 0)


def test_rate_manager_export():
    tdir, h5path = setuph5()
    rmg = RateManager(h5path)
    rmg.export_training_set(tdir)

    ss = rmg.samples[0]
    feats = IndentationRater.get_feature_names()

    for ff, si in zip(feats, ss):
        fi = np.loadtxt(tdir / "train_{}.txt".format(ff))
        # :.2e, because features are not stored with high accuracy
        assert np.ndarray.item(fi) == float("{:.2e}".format(si))


def test_rate_manager_import():
    pin = data_path / "rate-export_1.7.8.h5"
    tdir = pathlib.Path(tempfile.mkdtemp(prefix="nanite_rate_load_"))
    h5path = tdir / pin.name
    shutil.copy2(pin, h5path)

    datalist = load_hdf5(h5path)
    rating = datalist[0]
    dataset = rating["data_set"]
    assert rating["rating"] == 5
    assert dataset.fit_properties["segment"] == "approach"
    # trigger recomputation of fit to force resetting of "segment"
    dataset.fit_model(range_x=[1.75e-5, 2.3e-5])
    assert dataset.fit_properties["segment"] == 0


def test_rate_manager_get_ts():
    path = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"
    tdir, h5path = setuph5(path=path)
    rmg = RateManager(h5path)
    x2, _ = rmg.get_training_set(which_type="binary")
    assert np.all(x2 == 1)
    x3, _ = rmg.get_training_set(which_type="continuous",
                                 prefilter_binary=True)
    x4, _ = rmg.get_training_set(remove_nans=True)
    assert np.all(np.hstack((x2, x3)) == x4)


@pytest.mark.filterwarnings('ignore::RuntimeWarning')
def test_rate_manager_get_ts_bad():
    path = data_path / "fmt-jpk-fd_map_bad_2013-05-27_2.jpk-force-map"
    tdir, h5path = setuph5(path=path)
    rmg = RateManager(h5path)
    x2, _ = rmg.get_training_set(which_type="binary")
    assert np.allclose(x2.flatten(), [1, 0, 1])
    x3, _ = rmg.get_training_set(which_type="continuous",
                                 prefilter_binary=True)
    assert x3.size == 0
    x4, _ = rmg.get_training_set(remove_nans=True)
    assert x4.size == 0


def test_rate_manager_get_ts_single():
    tdir, h5path = setuph5()
    rmg = RateManager(h5path)
    x2, _ = rmg.get_training_set(which_type="binary")
    assert np.all(x2 == 1)
    x3, _ = rmg.get_training_set(which_type="continuous",
                                 prefilter_binary=True)
    x4, _ = rmg.get_training_set(remove_nans=True)
    assert np.all(np.hstack((x2, x3)) == x4)


def test_write():
    tdir, h5path, idnt = setuph5(ret_idnt=True)

    with h5py.File(str(h5path), mode="r") as hi:
        # experimental data
        assert "4443b7" in hi["data"]
        # a few attributes
        attrs = hi["analysis/4443b7_0"].attrs
        assert attrs["fit model_key"] == "hertz_para"
        assert not attrs["fit optimal_fit_edelta"]
        assert attrs["fit preprocessing"] == "compute_tip_position"
        assert np.allclose(hi["analysis/4443b7_0"]["fit"], idnt["fit"],
                           equal_nan=True)


def test_write_read():
    tdir, h5path = setuph5()

    datalist = load_hdf5(h5path)
    assert datalist[0]["rating"] == 5


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import warnings

from . import model_conical_indenter  # noqa: F401
from . import model_hertz_paraboloidal  # noqa: F401
from . import model_hertz_three_sided_pyramid  # noqa: F401
from . import model_sneddon_spherical  # noqa: F401
from . import model_sneddon_spherical_approximation  # noqa: F401

from .core import NaniteFitModel  # noqa: F401
from . import residuals  # noqa: F401
from .logic import models_available, register_model
from .logic import deregister_model, load_model_from_file  # noqa: F401


def compute_anc_parms(idnt, model_key):
    """Compute ancillary parameters for a force-distance dataset

    Ancillary parameters include parameters that:

    - are unrelated to fitting: They may just be important parameters
      to the user.
    - require the entire dataset: They cannot be extracted during
      fitting, because they require more than just the approach xor
      retract curve to compute (e.g. hysteresis, jump of retract curve
      at maximum indentation). They may, additionally, depend on
      initial fit parameters set by the user.
    - require a fit: They are dependent on fitting parameters but
      are not required during fitting.

    Notes
    -----
    If an ancillary parameter name matches that of a fitting parameter,
    then it is assumed that it can be used for fitting. Please see
    :func:`nanite.indent.Indentation.get_initial_fit_parameters`
    and :func:`nanite.fit.guess_initial_parameters`.

    Ancillary parameters are set to `np.nan` if they cannot be
    computed.

    Parameters
    ----------
    idnt: nanite.indent.Indentation
        The force-distance data for which to compute the ancillary
        parameters
    model_key: str
        Name of the model

    Returns
    -------
    ancillaries: collections.OrderedDict
        key-value dictionary of ancillary parameters
    """
    md = models_available[model_key]
    return md.compute_ancillaries(idnt)


def get_anc_parms(idnt, model_key):
    warnings.warn(
        "Method `get_anc_parms` is deprecated. Please use "
        + "`compute_anc_parms` instead.",
        DeprecationWarning)
    return compute_anc_parms(idnt, model_key)


def get_anc_parm_keys(model_key):
    """Return the key names of a model's ancillary parameters"""
    md = models_available[model_key]
    return md.get_anc_parm_keys()


def get_init_parms(model_key):
    """Get initial fit parameters for a model"""
    md = models_available[model_key]
    return md.get_parameter_defaults()


def get_model_by_name(name):
    """Convenience function to obtain a model by name instead of by key"""
    for key in models_available:
        if models_available[key].model_name == name:
            return models_available[key]
    else:
        raise KeyError("No model with name '{}'!".format(name))


def get_parm_name(model_key, parm_key):
    """Return parameter label

    Parameters
    ----------
    model_key: str
        The model key (e.g. "hertz_cone")
    parm_key: str
        The parameter key (e.g. "E")

    Returns
    -------
    parm_name: str
        The parameter label (e.g. "Young's Modulus")
    """
    md = models_available[model_key]
    return md.get_parm_name(parm_key)


def get_parm_unit(model_key, parm_key):
    """Return parameter unit

    Parameters
    ----------
    model_key: str
        The model key (e.g. "hertz_cone")
    parm_key: str
        The parameter key (e.g. "E")

    Returns
    -------
    parm_unit: str
        The parameter unit (e.g. "Pa")
    """
    md = models_available[model_key]
    return md.get_parm_unit(parm_key)


# Populate list of available fit models
_loc = locals().copy()
for _item in _loc:
    if _item.startswith("model_"):
        register_model(_loc[_item])



"""Test of data set functionalities"""
import pathlib

import numpy as np
import pytest

from nanite import IndentationGroup


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


@pytest.mark.filterwarnings('ignore::nanite.smooth.'
                            + 'DoubledSmoothingWindowWarning')
def test_app_ret():
    grp = IndentationGroup(jpkfile)
    idnt = grp[0]
    idnt.apply_preprocessing(["compute_tip_position"])
    height = np.array(idnt["height (measured)"], copy=True)
    tip_position = np.array(idnt["tip position"], copy=True)
    idnt.apply_preprocessing(["smooth_height"])
    hms = np.array(idnt["height (measured)"])
    assert not np.all(height == hms)
    idnt.apply_preprocessing(["compute_tip_position",
                              "smooth_height"])
    hms2 = np.array(idnt["height (measured)"])
    assert np.all(hms == hms2)
    assert not np.all(idnt["tip position"] == tip_position)


def test_tip_sample_separation():
    grp = IndentationGroup(jpkfile)
    idnt = grp[0]
    # This computation correctly reproduces the column
    # "Vertical Tip Position" as it is exported by the
    # JPK analysis software with the checked option
    # "Use Unsmoothed Height".
    idnt.apply_preprocessing(["compute_tip_position"])
    tip = np.array(idnt["tip position"])
    assert tip[0] == 2.2803841798545836e-05


def test_correct_app_ret():
    grp = IndentationGroup(jpkfile)
    idnt = grp[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_split_approach_retract"])
    a = idnt["segment"][idnt["segment"] == 0]
    assert len(a) == 2006


def test_correct_force_offset():
    grp = IndentationGroup(jpkfile)
    idnt = grp[0]
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset"])
    idp = idnt.estimate_contact_point_index()
    assert np.allclose(np.average(idnt["force"][:idp]), 0)


@pytest.mark.parametrize(
    "metadata,software",
    [
        (None, "JPK"),
        ({"software": "custom1a"}, "custom1a"),
    ])
def test_metadata_override(metadata, software):
    grp = IndentationGroup(jpkfile, meta_override=metadata)
    idnt = grp[0]
    assert idnt.metadata["software"] == software



from os.path import dirname, exists, realpath
from setuptools import setup, Extension, find_packages
import sys

# numpy and cython are installed via pyproject.toml [build-system]
import numpy as np

maintainer = "Paul Müller"
maintainer_email = "dev@craban.de"
description = 'Loading, fitting, and rating AFM force-distance data'
name = 'nanite'
year = "2018"


sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
try:
    from _version import version  # noqa: F821
except BaseException:
    version = "unknown"


setup(
    name=name,
    maintainer=maintainer,
    maintainer_email=maintainer_email,
    url='https://github.com/AFM-analysis/nanite',
    version=version,
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license="GPL v3",
    description=description,
    long_description=open('README.rst').read() if exists('README.rst') else '',
    install_requires=["afmformats>=0.16.4",
                      "h5py>=2.8.0",
                      "lmfit>=1",
                      "numpy>=1.16.0",  # cython build
                      "scikit-learn>=0.23.0",  # rating tests
                      "scipy",
                      ],
    ext_modules=[Extension("nanite.model.model_sneddon_spherical",
                           sources=[
                               "nanite/model/model_sneddon_spherical.pyx"],
                           include_dirs=[np.get_include()],
                           )
                 ],
    extras_require={
        'CLI':  ["appdirs",
                 "matplotlib>=2.2.2",
                 "tifffile>=0.15.0",
                 ],
    },
    python_requires='>=3.6, <4',
    entry_points={
        "console_scripts": [
            "nanite-setup-profile = nanite.cli:setup_profile [CLI]",
            "nanite-rate = nanite.cli:rate [CLI]",
            "nanite-fit = nanite.cli:fit [CLI]",
            "nanite-generate-training-set = nanite.cli:generate_training_set"
            + " [CLI]",
        ],
    },
    keywords=["atomic force microscopy",
              "mechanical phenotyping",
              "tissue analysis"],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Scientific/Engineering :: Visualization',
        'Intended Audience :: Science/Research'
    ],
    platforms=['ALL'],
)



"""Test of data set functionalities"""
import pathlib
import tempfile

import numpy as np
import pytest

import afmformats
import nanite
import nanite.model


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_apply_preprocessing():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    # apply preprocessing by manually setting the list
    idnt.preprocessing = ["compute_tip_position"]
    idnt.apply_preprocessing()


def test_apply_preprocessing_remember_fit_properties():
    """
    Normally, the fit properties would be overridden
    if the preprocessing changes. For user convenience,
    nanite remembers it. This is the test
    """
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    idnt.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    cp1 = 1.8029310065572342e-05
    inparams["contact_point"].set(cp1)
    inparams["contact_point"].vary = False

    # Fit with absolute full range
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(0, 0),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    assert cp1 == idnt.fit_properties["params_initial"]["contact_point"].value
    assert cp1 == idnt.fit_properties["params_fitted"]["contact_point"].value

    # Change preprocessing
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_force_offset"])
    assert cp1 == idnt.fit_properties["params_initial"]["contact_point"].value


def test_basic():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    # tip-sample separation
    idnt.apply_preprocessing(["compute_tip_position"])
    assert idnt.preprocessing == ["compute_tip_position"]
    assert idnt["tip position"][0] == 2.2803841798545836e-05
    # correct for an offset in the tip
    idnt.apply_preprocessing(["compute_tip_position",
                              "correct_tip_offset"])
    # This value is subject to change if a better way to estimate the
    # contact point is found:
    assert idnt["tip position"][0] == 4.765854684370548e-06


def test_export():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    # tip-sample separation
    idnt.apply_preprocessing(["compute_tip_position"])
    # create temporary file
    _, path = tempfile.mkstemp(suffix=".tab", prefix="nanite_idnt_export")
    idnt.export_data(path)
    data = afmformats.load_data(path)[0]
    assert len(data) == 4000
    assert np.allclose(data["force"][100], -4.853736717639109e-10)
    assert np.allclose(data["height (measured)"][100], 2.256791903750211e-05,
                       atol=1e-10, rtol=0)
    assert data["segment"][100] == 0
    assert np.allclose(data["time"][100], 0.04999999999999999)
    assert np.allclose(data["tip position"][100], 2.255675939721752e-05,
                       atol=1e-10, rtol=0)
    assert data["segment"][3000] == 1


def test_fitting():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    idnt.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8e-5)

    # Fit with absolute full range
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(0, 0),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    params = idnt.fit_properties["params_fitted"]
    assert np.allclose(params["contact_point"].value, 1.8029310201193193e-05)
    assert np.allclose(params["E"].value, 14741.958242422093)

    # Fit with absolute short
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(17e-06, 19e-6),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    params2 = idnt.fit_properties["params_fitted"]
    assert np.allclose(params2["contact_point"].value, 1.8028461828272924e-05)
    assert np.allclose(params2["E"].value, 14840.840404880484)

    # Fit with relative to initial fit
    idnt.fit_model(model_key="hertz_para",
                   params_initial=params2,
                   range_x=(-2e-6, 1e-6),
                   range_type="relative cp",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    params3 = idnt.fit_properties["params_fitted"]
    # These results are subject to change if the "relative cp" method is
    # changed.
    assert np.allclose(params3["contact_point"].value, 1.8028478083499856e-05)
    assert np.allclose(params3["E"].value, 14839.821714634612)


@pytest.mark.filterwarnings('ignore::nanite.fit.FitWarning')
def test_get_initial_fit_parameters():
    """This is a convenience function"""
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    # A: sanity check
    fp = idnt.get_initial_fit_parameters()
    assert fp["contact_point"].value == 0, "need to get tip position first"
    # B: get default fit parameters (hertz_para)
    idnt.apply_preprocessing(["compute_tip_position"])
    fp = idnt.get_initial_fit_parameters()
    for kk in ['E', 'R', 'nu', 'contact_point', 'baseline']:
        assert kk in fp.keys()
    # C: set other model and get fit parameters
    idnt.fit_properties["model_key"] = "hertz_pyr3s"
    fp2 = idnt.get_initial_fit_parameters()
    for kk in ['E', 'alpha', 'nu', 'contact_point', 'baseline']:
        assert kk in fp2.keys()
    # D: fit and get from fit_properties
    idnt.fit_model(params_initial=fp2)
    fp3 = idnt.get_initial_fit_parameters()
    assert fp2 == fp3


def test_get_model():
    md = nanite.model.model_hertz_paraboloidal
    model_name = "parabolic indenter (Hertz)"
    md2 = nanite.model.get_model_by_name(model_name).module
    assert md2 is md


def test_preprocessing_reset():
    fd = nanite.IndentationGroup(jpkfile)[0]
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"],
                           options={"correct_tip_offset": {
                               "method": "fit_constant_line"}})

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8e-5)

    # Perform fit and make sure that all properties are set
    fd.fit_model(model_key="hertz_para",
                 params_initial=inparams,
                 range_x=(0, 0),
                 range_type="absolute",
                 x_axis="tip position",
                 y_axis="force",
                 segment="approach",
                 weight_cp=False)
    fd.rate_quality(training_set="zef18",
                    regressor="Extra Trees")
    assert fd._rating is not None
    assert fd.preprocessing == ["compute_tip_position",
                                "correct_force_offset",
                                "correct_tip_offset"]
    assert fd.preprocessing_options == {"correct_tip_offset": {
                                        "method": "fit_constant_line"}}
    assert not fd._preprocessing_details
    assert np.allclose(
        fd._fit_properties["params_fitted"]["contact_point"].value,
        -3.5147555568064786e-06,
        atol=0)
    assert "tip position" in fd

    # Change preprocessing and make sure properties are reset
    fd.apply_preprocessing(["compute_tip_position",
                            "correct_force_offset",
                            "correct_tip_offset"],
                           options={"correct_tip_offset": {
                               "method": "fit_constant_polynomial"}})
    assert fd._rating is None
    assert fd.preprocessing == ["compute_tip_position",
                                "correct_force_offset",
                                "correct_tip_offset"]
    assert fd.preprocessing_options == {"correct_tip_offset": {
                                        "method": "fit_constant_polynomial"}}
    assert not fd._preprocessing_details
    assert "params_fitted" not in fd._fit_properties
    assert "tip position" in fd

    # Change preprocessing, removing tip position
    fd.apply_preprocessing(["correct_force_offset"])
    assert "tip position" not in fd


def test_rate_quality_cache():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    idnt.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8e-5)

    # Fit with absolute full range
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(0, 0),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)
    r1 = idnt.rate_quality(training_set="zef18",
                           regressor="Extra Trees")
    assert idnt._rating[-1] == r1
    r2 = idnt.rate_quality(training_set="zef18",
                           regressor="Extra Trees")
    assert r1 == r2


def test_rate_quality_disabled():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    idnt.apply_preprocessing(["compute_tip_position"])

    inparams = nanite.model.model_hertz_paraboloidal.get_parameter_defaults()
    inparams["baseline"].vary = True
    inparams["contact_point"].set(1.8e-5)

    # Fit with absolute full range
    idnt.fit_model(model_key="hertz_para",
                   params_initial=inparams,
                   range_x=(0, 0),
                   range_type="absolute",
                   x_axis="tip position",
                   y_axis="force",
                   segment="approach",
                   weight_cp=False)

    r1 = idnt.rate_quality(training_set="zef18",
                           regressor="none")
    assert r1 == -1


def test_rate_quality_nofit():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    r1 = idnt.rate_quality()
    assert r1 == -1


def test_repr_str():
    ds1 = nanite.IndentationGroup(jpkfile)
    idnt = ds1[0]
    assert "AFMForceDistance" not in str(idnt)
    assert "Indentation" in str(idnt)
    assert "fmt-jpk-fd_spot3-0192.jpk-force" in str(idnt)
    assert "Indentation" in repr(idnt)
    assert "fmt-jpk-fd_spot3-0192.jpk-force" in repr(idnt)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import lmfit
import numpy as np


def get_parameter_defaults():
    """Return the default model parameters"""
    # The order of the parameters must match the order
    # of ´parameter_names´ and ´parameter_keys´.
    params = lmfit.Parameters()
    params.add("E", value=3e3, min=0)
    params.add("R", value=10e-6, min=0, vary=False)
    params.add("nu", value=.5, min=0, max=0.5, vary=False)
    params.add("contact_point", value=0)
    params.add("baseline", value=0)
    return params


def hertz_paraboloidal(delta, E, R, nu, contact_point=0, baseline=0):
    r"""Hertz model for a paraboloidal indenter

    .. math::

        F = \frac{4}{3}
            \frac{E}{1-\nu^2}
            \sqrt{R}
            \delta^{3/2}

    Parameters
    ----------
    delta: 1d ndarray
        Indentation [m]
    E: float
        Young's modulus [N/m²]
    R: float
        Tip radius [m]
    nu: float
        Poisson's ratio
    contact_point: float
        Indentation offset [m]
    baseline: float
        Force offset [N]

    Returns
    -------
    F: float
        Force [N]

    Notes
    -----
    The derivation in :cite:`Sneddon1965` reads

    .. math::

        F = \frac{4}{3}
            \frac{E}{1-\nu^2}
            \sqrt{2k}
            \delta^{3/2},

    where :math:`k` is defined by the paraboloid equation

    .. math::

        \rho^2 = 4kz.

    As follows from the derivations in :cite:`LandauLifshitz`, this
    model is valid for either of the two cases:

    - Indentation of a plane (infinite radius) with Young's modulus
      :math:`E` by a sphere with infinite Young's modulus and radius
      :math:`R`, or
    - Indentation of a sphere with radius :math:`R` and Young's modulus
      :math:`E` by a plane (infinite radius) with infinite Young's modulus.


    These approximations are made by the Hertz model:

    - The sample is isotropic.
    - The sample is a linear elastic solid.
    - The sample is extended infinitely in one half space.
    - The indenter is not deformable.
    - There are no additional interactions between sample and indenter.

    Additional assumptions:

    - no surface forces
    - If the indenter is spherical, then its radius :math:`R` is much
      larger than the indentation depth :math:`\delta`.

    References
    ----------
    Sneddon (1965) :cite:`Sneddon1965` (equation 6.9),
    Theory of Elasticity by Landau and Lifshitz (1959)
    :cite:`LandauLifshitz` (§9 Solid bodies in contact, equation 9.14)
    """
    aa = 4/3 * E/(1-nu**2)*np.sqrt(R)
    root = contact_point-delta
    pos = root > 0
    bb = np.zeros_like(delta)
    bb[pos] = (root[pos])**(3/2)
    return aa*bb + baseline


model_doc = hertz_paraboloidal.__doc__
model_func = hertz_paraboloidal
model_key = "hertz_para"
model_name = "parabolic indenter (Hertz)"
parameter_keys = ["E", "R", "nu", "contact_point", "baseline"]
parameter_names = ["Young's Modulus", "Tip Radius",
                   "Poisson's Ratio", "Contact Point", "Force Baseline"]
parameter_units = ["Pa", "m", "", "m", "N"]
valid_axes_x = ["tip position"]
valid_axes_y = ["force"]



"""Test group functionalities"""
import pathlib
import tempfile
import shutil

from afmformats.errors import MissingMetaDataError
import pytest

from nanite import Indentation, IndentationGroup, load_group


data_path = pathlib.Path(__file__).resolve().parent / "data"
jpkfile = data_path / "fmt-jpk-fd_spot3-0192.jpk-force"


def test_base():
    ds1 = IndentationGroup(jpkfile)
    ds2 = IndentationGroup(jpkfile)

    ds3 = ds1 + ds2
    assert len(ds3) == 2
    assert len(ds2) == 1
    assert len(ds1) == 1

    ds2 += ds3
    assert len(ds3) == 2
    assert len(ds2) == 3

    for apret in ds3:
        assert isinstance(apret, Indentation)
    # test repr
    print(ds3)


def test_subgroup():
    tmp = tempfile.mkdtemp(prefix="test_nanite_group")
    shutil.copy(
        data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map", tmp)
    shutil.copy(data_path / "fmt-jpk-fd_map2x2_extracted.jpk-force-map", tmp)
    shutil.copy(
        data_path / "fmt-jpk-fd_flipsign_2015.05.22-15.31.49.352.jpk-force",
        tmp)

    grp = load_group(tmp)
    exp = pathlib.Path(tmp) / "fmt-jpk-fd_map2x2_extracted.jpk-force-map"
    subgrp = grp.subgroup_with_path(path=exp)
    assert len(grp) == 8
    assert len(subgrp) == 4
    assert subgrp[0].path == exp


def test_open_dat_without_spring_constant():
    # There are data files without spring constant but with force in nN.
    # nanite requires the spring constant for computation of the tip
    # position which we check for here.
    tmp = tempfile.mkdtemp(prefix="test_nanite_group")
    shutil.copy(data_path / "fmt-afm-workshop-fd_single_2021-10-22_14.16.csv",
                tmp)

    # try without explicit metadata
    with pytest.raises(MissingMetaDataError, match="spring constant"):
        load_group(tmp)

    # try with metadata
    grp = load_group(tmp, meta_override={"spring constant": 20})
    assert grp[0].metadata["spring constant"] == 20


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



"""Test usage of user-defined training set"""
import pathlib
import tempfile

import numpy as np

from nanite import IndentationGroup
from nanite.rate import IndentationRater


data_path = pathlib.Path(__file__).parent / "data"
jpkfile = data_path / "fmt-jpk-fd_map-data-reference-points.jpk-force-map"


def setup_training_set(n=300):
    tdir = tempfile.mkdtemp(prefix="test_nanite_rate_ts_")
    tdir = pathlib.Path(tdir)
    np.random.set_state(np.random.RandomState(47).get_state())
    for bb in IndentationRater.get_feature_names(which_type="binary"):
        bvals = np.random.choice([0, 1], size=n, p=[.05, .95])
        np.savetxt(tdir / "train_{}.txt".format(bb), bvals)
    for cc in IndentationRater.get_feature_names(which_type="continuous"):
        cvals = np.random.random_sample(size=n)
        np.savetxt(tdir / "train_{}.txt".format(cc), cvals)
    rating = np.random.choice(range(11), size=n)
    np.savetxt(tdir / "train_response.txt", rating)
    return tdir


def test_user_training_set():
    tdir = setup_training_set()
    # load a curve
    idnt = IndentationGroup(jpkfile)[0]
    # fit it
    idnt.fit_model(model_key="sneddon_spher_approx",
                   preprocessing=["compute_tip_position",
                                  "correct_force_offset"])
    r1 = idnt.rate_quality(regressor="Extra Trees", training_set="zef18")
    assert r1 > 9, "sanity check"
    r2 = idnt.rate_quality(regressor="Extra Trees", training_set=tdir)
    assert 4 < r2 < 5, "with the given random state we end up at 4.55"


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



import warnings

import afmformats
from .indent import Indentation

#: The default imaging modality when loading AFM data. Set this to `None`
#: to also be able to load e.g. creep-compliance data. See issue
#: https://github.com/AFM-analysis/nanite/issues/11 for more
#: information. Note that especially the export of rating containers
#: may not work with any imaging modality other than force-distance.
DEFAULT_MODALITY = "force-distance"


def get_data_paths(path):
    """Return list of data paths with force-distance data

    DEPRECATED
    """
    warnings.warn("`get_data_paths` is deprecated! Please use "
                  + "afmformats.find_data(path, modality='force-distance') "
                  + "instead!",
                  DeprecationWarning)
    return afmformats.find_data(path, modality=DEFAULT_MODALITY)


def get_data_paths_enum(path, skip_errors=False):
    """Return a list with paths and their internal enumeration

    Parameters
    ----------
    path: str or pathlib.Path or list of str or list of pathlib.Path
        path to data files or directory containing data files;
        if directories are given, they are searched recursively
    skip_errors: bool
        skip paths that raise errors

    Returns
    -------
    path_enum: list of lists
        each entry in the list is a list of [pathlib.Path, int],
        enumerating all curves in each file
    """
    paths = afmformats.find_data(path, modality=DEFAULT_MODALITY)
    enumpaths = []
    for pp in paths:
        try:
            data = load_data(pp)
        except BaseException:
            if skip_errors:
                continue
            else:
                raise
        for dd in data:
            enumpaths.append([pp, dd.enum])
    return enumpaths


def get_load_data_modality_kwargs():
    """Return imaging modality kwargs for afmformats.load_data

    Uses :const:`DEFAULT_MODALITY`.

    Returns
    -------
    kwargs: dict
        keyword arguments for :func:`afmformats.load_data`
    """
    kwargs = {
        "modality": DEFAULT_MODALITY,
        "data_classes_by_modality": {
            "force-distance": Indentation,
            "creep-compliance": Indentation,
            "stress-relaxation": Indentation}
    }
    return kwargs


def load_data(path, callback=None, meta_override=None):
    """Load data and return list of :class:`afmformats.AFMForceDistance`

    This is essentially a wrapper around
    :func:`afmformats.formats.find_data` and
    :func:`afmformats.formats.load_data` that returns
    force-distance datasets.

    Parameters
    ----------
    path: str or pathlib.Path or list of str or list of pathlib.Path
        path to data files or directory containing data files;
        if directories are given, they are searched recursively
    callback: callable
        function for progress tracking; must accept a float in
        [0, 1] as an argument.
    meta_override: dict
        if specified, contains key-value pairs of metadata that
        are used when loading the files
        (see :data:`afmformats.meta.META_FIELDS`)
    """
    paths = afmformats.find_data(path, modality=DEFAULT_MODALITY)
    data = []
    for ii, pp in enumerate(paths):
        measurements = afmformats.load_data(
            pp,
            # recurse callback function with None as default
            callback=lambda x: callback((ii + x) / len(paths))
            if callback else None,
            meta_override=meta_override,
            **get_load_data_modality_kwargs()
        )
        data += measurements
    return data



"""Test NaniteFitModel class"""
import pathlib

import numpy as np
import pytest

import nanite.model.core

from common import MockModelModule

data_path = pathlib.Path(__file__).parent / "data"


def test_compute_anc_max_indent():
    ds1 = nanite.IndentationGroup(
        data_path / "fmt-jpk-fd_spot3-0192.jpk-force")
    fd = ds1[0]
    # correct for an offset in the tip
    fd.apply_preprocessing(preprocessing=["compute_tip_position",
                                          "correct_tip_offset"],
                           options={
                               "correct_tip_offset": {
                                   "method": "deviation_from_baseline"
                               }
                           })
    # fit a model
    fd.fit_model(model_key="hertz_para")
    # get the ancillary parameter
    max_indent = nanite.model.core.compute_anc_max_indent(fd)
    assert np.allclose(max_indent, 1.2950734601921855e-07, atol=0, rtol=1e-8)


def test_model_bad_incomplete():
    class BadModel:
        model_key = "peterpan"
    bad_mod = BadModel()
    with pytest.raises(nanite.model.core.ModelIncompleteError,
                       match="parameter_names"):
        nanite.model.core.NaniteFitModel(bad_mod)


def test_model_bad_incomplete_anc():
    bad_mod = MockModelModule(model_key="peterpan",
                              compute_ancillaries=lambda x: {"peter": 1.2})
    with pytest.raises(nanite.model.core.ModelIncompleteError,
                       match="parameter_anc_keys"):
        nanite.model.core.NaniteFitModel(bad_mod)


def test_model_bad_labels_not_unique():
    parameter_names_wrong = [
        "SAME", "SAME", "Two", "Three", "Four"]
    bad_mod = MockModelModule(model_key="peterpan",
                              parameter_names=parameter_names_wrong)
    with pytest.raises(nanite.model.core.ModelImplementationError,
                       match="'parameter_names' should be unique for"):
        nanite.model.core.NaniteFitModel(bad_mod)


def test_model_bad_spaces_in_units():
    parameter_units_wrong = ["Pa", "m", " ", "m", "N"]
    bad_mod = MockModelModule(model_key="peterpan",
                              parameter_units=parameter_units_wrong)
    with pytest.warns(nanite.model.core.ModelImplementationWarning,
                      match="should not contain leading or trailing spaces"):
        nanite.model.core.NaniteFitModel(bad_mod)


def test_model_bad_spaces_in_units_anc():
    bad_mod = MockModelModule(
                compute_ancillaries=lambda x: {"E": np.nan},
                parameter_anc_keys=["E"],
                parameter_anc_names=["ancillary E guess"],
                parameter_anc_units=["Pa "],
                model_key="test2")
    with pytest.warns(nanite.model.core.ModelImplementationWarning,
                      match="should not contain leading or trailing spaces"):
        nanite.model.core.NaniteFitModel(bad_mod)


def test_model_get_anc_parm_keys():
    with MockModelModule(model_key="peterpan") as md:
        akeys = md.get_anc_parm_keys()
        assert "max_indent" in akeys


def test_model_get_anc_parm_keys_2():
    with MockModelModule(model_key="peterpan",
                         compute_ancillaries=lambda x: {"hans": 1.2},
                         parameter_anc_keys=["hans"],
                         parameter_anc_names=["Hans"],
                         parameter_anc_units=["N"],
                         ) as md:
        akeys = md.get_anc_parm_keys()
        assert "max_indent" in akeys
        assert "hans" in akeys


def test_model_get_parm_name():
    with MockModelModule(model_key="peterpan",
                         compute_ancillaries=lambda x: {"hans": 1.2},
                         parameter_anc_keys=["hans"],
                         parameter_anc_names=["Hans"],
                         parameter_anc_units=["N"],
                         ) as md:
        assert md.get_parm_name("E") == "Young's Modulus"
        assert md.get_parm_name("hans") == "Hans"
        assert md.get_parm_name("max_indent") == "Maximum indentation"
        with pytest.raises(KeyError, match="Could not find parameter name"):
            md.get_parm_name("peter")


def test_model_get_parm_unit():
    with MockModelModule(model_key="peterpan",
                         compute_ancillaries=lambda x: {"hans": 1.2},
                         parameter_anc_keys=["hans"],
                         parameter_anc_names=["Hans"],
                         parameter_anc_units=["N"],
                         ) as md:
        assert md.get_parm_unit("E") == "Pa"
        assert md.get_parm_unit("hans") == "N"
        assert md.get_parm_unit("max_indent") == "m"
        with pytest.raises(KeyError, match="Could not find parameter unit"):
            md.get_parm_unit("peter")


def test_model_repr_str():
    with MockModelModule(model_key="peterpan") as md:
        assert "peterpan" in repr(md)
        assert "NaniteFitModel" in repr(md)
        assert "peterpan" in str(md)
        assert "NaniteFitModel" in str(md)



"""Test smoothing functionalities"""
import numpy as np
import pytest

from nanite import smooth


def test_smooth_monotone():
    x = np.array([0]*30
                 + [1]*30
                 + [2]*30,
                 dtype=float)
    sm = smooth.smooth_axis_monotone(data=x, window=5)
    assert np.unique(sm).size == sm.size


@pytest.mark.filterwarnings('ignore::nanite.smooth.'
                            + 'DoubledSmoothingWindowWarning')
def test_smooth_monotone2():
    x = np.array([0]*30
                 + [1]*30
                 + [0]*10  # the algorithm will smooth over this part
                 + [1]*30,
                 dtype=float)
    sm = smooth.smooth_axis_monotone(data=x, window=5)
    assert np.unique(sm).size == sm.size


@pytest.mark.filterwarnings('ignore::nanite.smooth.'
                            + 'DoubledSmoothingWindowWarning')
def test_smooth_monotone_maxiter():
    # this is not solvable
    x = np.array([0]*30
                 + [1]*30
                 + [0]*30,
                 dtype=float)
    try:
        smooth.smooth_axis_monotone(data=x, window=5, max_iter=100)
    except ValueError:
        pass
    else:
        assert False


@pytest.mark.filterwarnings('ignore::nanite.smooth.'
                            + 'DoubledSmoothingWindowWarning')
def test_smooth_monotone_maxiter2():
    # this is not solvable
    x = np.array([0]*30
                 + [1]*30
                 + [0]*30,
                 dtype=float)
    try:
        smooth.smooth_axis_monotone(data=x, window=5,
                                    max_iter=0  # stop in first loop
                                    )
    except ValueError:
        pass
    else:
        assert False


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()
